/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Code Notes:
//     Chose to put everything, even errors, on stdout; otherwise see weird
//     and confusing inversions in the order of print-outs when errors occur
//     and are printed before their antecedents.
// TODO (?) Add verbosity handling?
// #############################################################################

// ============================================================================
// #inclues; shared_addresses is shared with ticket kernel.
// ============================================================================
#include "admin.h"
#include "../common/shared_addresses.c"

// ============================================================================
// Globals
// ============================================================================
struct name_mapping_raw map_addrs_raw[] = HOST_ADDRESSES_TABLE;
struct name_mapping     *map_addrs;
// info on packet to send
char admin_packet[ADMIN_PACKET_SIZE];
// info on how to send it.
struct name_mapping source_addr, dest_addr;
// which interface should send the packet.
char *interface;
// what i'm doing. 
unsigned int mode;


// ============================================================================
// Create the kill packet (ethernet specific)
//
// The box to commit suicide only checks for tos 0x9999 and '<sop>DIE!'.
// ============================================================================
unsigned int makemsg(char *msg) {
	int len;

	len = strlen(msg);
	// forge IP header right here.... (plop).
	memcpy(&(admin_packet[ETH_HDR_LEN]), msg, len);

	return(ETH_HDR_LEN+len);
}


// ============================================================================
// Drop root priveleges by becoming 'nobody'.
// ============================================================================
void drop_priv(void) {
	struct passwd *pw = NULL;

	// printf("Current UID %d, EUID %d, GID %d, EGID %d\n", getuid(), geteuid(), getgid(), getegid());
	// get password data for the person we're to become...
	if (!(pw = getpwnam("nobody"))) {
		printf("User 'nobody' doesn't exist!!!\n");
		exit(1);
	}

	// drop stuff in the right order.
	if (!initgroups("nobody", pw->pw_gid)) {
		if (!setgid(pw->pw_gid)) {
			if (!setuid(pw->pw_uid)) {
				// success!
				// printf("After priv drop UID %d, EUID %d, GID %d, EGID %d\n", getuid(), geteuid(), getgid(), getegid());
				return;
			}
		}
	}
	printf("Failed to change program owership to 'nobody' (UID %d, GID %d)\n", pw->pw_uid, pw->pw_gid);
	exit(1);
}


// ============================================================================
// Send the kill message to this mac address.
// This code doesn't use the common network code,
//    because it does raw socket operations.
// ============================================================================
int send_admin(unsigned int packet_len) {
	int sockfd;
	struct sockaddr_ll my_addr;
	struct ifreq ireq;
	// for use with sendmsg
	struct msghdr send_hdr;
	struct iovec send_vec;

	// minimum ethernet length.
	if (packet_len<MIN_ETH_LEN) {
		packet_len = MIN_ETH_LEN;
	}

	// Get a raw socket file descriptor for all interfaces.
	// See packet(7), bind(2)
	printf("\tOpening socket... ");
	if ((sockfd = socket(PF_PACKET, SOCK_RAW, ETH_PROTOCOL))<0) {
		printf("\nsocket() call failed: %s\n", strerror(errno));
		return(0);
	}

	// First, get the index of interface given the name
	printf("addressing interface... ");
	memset(&ireq, 0, sizeof(ireq));
	(void)strncpy(ireq.ifr_name, interface, IFNAMSIZ);
	if (ioctl(sockfd, SIOCGIFINDEX, &ireq)) {
		printf("\nioctl() to get interface index in open_raw_socket failed: %s\n", strerror(errno));
		return(0);
	}

	// Now, we use the interface index to actually perform the binding operation.
	printf("binding... ");
	memset(&my_addr, 0, sizeof(struct sockaddr_ll));
	my_addr.sll_family = AF_PACKET;
	my_addr.sll_protocol = ETH_PROTOCOL;
	my_addr.sll_ifindex = ireq.ifr_ifindex;
	if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(my_addr))) {
		printf("\nbind() raw socket failed: %s\n", strerror(errno));
		return(0);
	}

	// drop root... rather late but still good practice.
	drop_priv();

	// set up the machine to send to...
	bzero(&send_hdr, sizeof(send_hdr));
	send_hdr.msg_iov = &send_vec;
	send_hdr.msg_iovlen = 1;
	bzero(&send_vec, sizeof(send_vec));
	send_vec.iov_base = admin_packet;
	send_vec.iov_len = packet_len;

	printf("sending... ");
	// send the packet.
	if (sendmsg(sockfd, &send_hdr, 0) != packet_len) {
		printf("\nwrite error to raw socket: %s\n", strerror(errno));
		return(0);
	}

	printf("sent.\n");

	return 1;
}


// ============================================================================
// Handle user permissions; you have to be root to get a raw socket.
// Will try to execute itself with 'sudo' if you haven't got root perms;
//     thus we need the argument list to do so properly.
// ============================================================================
int root_check(int argc, char *argv[]) {
	char *systembuf;
	int idx, len;

	// if root, we're good to go.
	if ((idx=getuid())==0)  {
		return 1;
	}

	printf("You need *root* access to run this program because it creates a raw socket.\n");

	// allocate room for new command-line
	len = 16;
	for (idx=0; idx<argc; idx++) {
		len+=strlen(argv[idx])+1;
	}
	if ((systembuf=(char *)malloc(len))==NULL) {
		printf("Failed to allocate string memory for system() call...\n");
		return 0;
	}
	
	// create new command line
	len = sprintf(systembuf, "sudo ");
	for (idx=0; idx<argc; idx++) {
		len += sprintf(&systembuf[len], "%s ", argv[idx]);
	}

	printf("\tI'll try to run myself via 'sudo' now, via `%s`...\n", systembuf);
	printf("\tYou could also just make me setuid root, but that would probably be dumb.\n\n");

	system(systembuf);

	// return here so we still give an error for user mistake.
	// plus, system's easier to call than the exec junk.
	return 0;
}


// ============================================================================
// How to use this program.
// ============================================================================
void usage(char *error) {
	char padstr1[16];
	char padstr2[16];
	int idx;

	if (error) { printf("\nUSAGE ERROR: %s\n", error); }
	printf("USAGE: sudo ./admin -m=[k|g|s] -s=[MAC|IP|Hostname] -d=[MAC|IP|Hosthame] <-i=eth0>\n");
	printf(" -m: mode, k=kill, g=get routes, s=set routes\n");
	printf(" -s: source to pretend I am\n");
	printf(" -d: destination to send to\n");
	printf(" -i: interface to send from (eth0 default)\n\n");
	printf("Known address mappings:\n");
	for (idx=0; map_addrs[idx].name!=NULL; idx++) {
		snprintf(padstr1, 16, "%s                ", map_addrs[idx].name);
		printf("\t%s   %s   %02X:%02X:%02X:%02X:%02X:%02X\n",
				padstr1,
				ip2str_nbo(map_addrs[idx].ip, padstr2),
				map_addrs[idx].mac[0], map_addrs[idx].mac[1],
				map_addrs[idx].mac[2], map_addrs[idx].mac[3],
				map_addrs[idx].mac[4], map_addrs[idx].mac[5]);
	}
	exit(1);
}


// ============================================================================
// Figure out what address has been passed in, and fill *addr with it.
// ============================================================================
int get_addr(char *str, struct name_mapping *addr) {
	int idx;

	// try and get arguments. these may fail, but that's OK.
	if (str2mac(str, addr->mac)) {
		// looks like a MAC. We should be good to go.
		// debug: printf("'%s' looks like a valid MAC address... No mapping required.\n", str);
		return 1;
	} else if (str2ip(str, &(addr->ip))) {
		// looks like an IP.
		for (idx=0; map_addrs[idx].name!=NULL; idx++) {
			if (map_addrs[idx].ip==addr->ip) {
				// debug: printf("'%s' looks looks like an IP... I know its mapping.\n", str);
				memcpy(addr, &(map_addrs[idx]), sizeof(struct name_mapping));
				return 1;
			}
		}

		printf("'%s' looks like an IP, but doesn't match any I know.\n", str);
	} else {
		// hostname?
		for (idx=0; map_addrs[idx].name!=NULL; idx++) {
			if (strcmp(map_addrs[idx].name, str)==0) {
				// debug: printf("'%s' looks like a hostname... I know its mapping.\n", str);
				memcpy(addr, &(map_addrs[idx]), sizeof(struct name_mapping));
				return 1;
			}
		}
		printf("'%s' looks like a hostname (doesn't parse as MAC/IP address), but I don't know it.\n", str);
	}
	return 0;
}


// ============================================================================
// return 1 if mac is null, 0 otherwise.
// ============================================================================
int nullmac(unsigned char *mac) {
	int idx;
	for (idx=0; idx<ETH_ALEN; idx++) {
		if (mac[idx]!='\0') {
			return 0;
		}
	}
	return 1;
}


// ============================================================================
// Initialize everything
// ============================================================================
void startup(int argc, char *argv[]) {
	int idx;

	// can do jack with no mapping
	if ((map_addrs = process_mapping(map_addrs_raw))==NULL) {
		printf("Bad mapping table in program...\n");
		if ((map_addrs = (struct name_mapping *)malloc(sizeof(struct name_mapping)))==NULL) {
			usage("Failed to create empty one too!\n");
		}
		map_addrs[0].name=NULL;
	}

	// initialize parameters
	mode=MODE_ERROR;
	memset(&source_addr, '\0', sizeof(struct name_mapping));
	memset(&dest_addr, '\0', sizeof(struct name_mapping));
	interface=NULL;
	// then get parameters.
	for (idx=1; idx<argc; idx++) {
		if (strlen(argv[idx])<4 || argv[idx][0]!='-' || argv[idx][2]!='=') {
			printf("Short argument ('%s')\n", argv[idx]);
			usage("Bad argument");
		}
		switch(argv[idx][1]) {
			case 'm': case 'M':
				// get mode.
				switch(argv[idx][3]) {
					case 'k': case 'K': mode=MODE_KILL; break;
					case 'g': case 'G': mode=MODE_GET;  break;
					case 's': case 'S': mode=MODE_SET;  break;
					default: usage("Mode must be 'k', 'g', or 's'!"); break;
				}
				break;
			case 's': case 'S':
				if (!get_addr(&(argv[idx][3]), &source_addr)) {
					usage("Bad source address");
				}
				break;
			case 'd': case 'D':
				if (!get_addr(&(argv[idx][3]), &dest_addr)) {
					usage("Bad dest address");
				}
				break;
			case 'i': case 'I':
				interface = &(argv[idx][3]);
				break;
			default:
				printf("Bad flag (%s, '%s') passed!\n", argv[idx], &(argv[idx][3]));
				usage("Bad flag passed.");
		}
	}
	// error checks.
	if (mode==MODE_ERROR) {
		usage("Pass in a valid mode!");
	}
	if (nullmac(source_addr.mac) || nullmac(dest_addr.mac)) {
		usage("Source or destination address is NULL!");
	}
	if (interface==NULL) {
		interface="eth0";
	}

	// you need to be root to run this (raw socket)
	// Do this LAST so that user can get more useful information first.
	if (!root_check(argc, argv)) {
		exit(1);
	}
}


// ============================================================================
// Main routine.
//  char tmpbuf[64];
//  printf("\tAbout to send kill message to %s:\n", mac2str(mac, tmpbuf));
// ============================================================================
int main(int argc, char *argv[]) {
	int retval;

	// initialize everything
	startup(argc, argv);

	// then do the work.
	bzero(admin_packet, ADMIN_PACKET_SIZE);
	memcpy(&(admin_packet[0*ETH_ALEN]),   dest_addr.mac, ETH_ALEN);
	memcpy(&(admin_packet[1*ETH_ALEN]), source_addr.mac, ETH_ALEN);
	admin_packet[2*ETH_ALEN+0] = (char)(((int)ETH_PROTOCOL)>>8);
	admin_packet[2*ETH_ALEN+1] = (char)(((int)ETH_PROTOCOL)&0xFF);

	switch(mode) {
		case MODE_KILL: retval=send_admin(makemsg("DIE!")); break;
		case MODE_SET:  retval=send_admin(makemsg("NEXT")); break;
		case MODE_GET:  retval=send_admin(makemsg("GET!")); break;
	}

	// send it out.
	if (!retval) {
		printf("Sending administrative message failed!\n");
	}

	// done!
	free_mapping(map_addrs);

	return 0;
}

