/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Master configuration file.
//
// This file contains all the statically-set configuration options...
//
// Modify with care.
// #############################################################################

#ifndef TICKET_CONFIG_H
#define TICKET_CONFIG_H

#include "shared_dump.h" /* Have to refer to sizeof(struct dump_packet) */

// =============================================================================
// USER LEVEL CONFIGURATION (Could wrap this stuff using __KERNEL__
// but not really necessary... and it makes the file harder to read).
// =============================================================================

// ------------------------------------------------------------
// Striping mechanism (writer.c)
//
// Split every 139,276 packets;
//   At (34*42)+8 bytes/packet=1436 bytes,
//   that's files of size (139,276*1,436 = 200,000,336B)
//
// I *HIGHLY* reccomend you leave this alone.
//     Larger files will leave larger spaces (and hence, wasted space)
//        at the end of disks given the striping logic.
//     Smaller files cause more overhead per file,
//        possibly leading to loss of packets.
//
// TODO: This changes if the struct dump_packet changes; make struct dump_packet
//       just be a char[68] to be libpcap-compatible?
// ------------------------------------------------------------
// How many packets to save per file, before splitting?
#define RECORDS_PER_FILE (200000336/sizeof(struct dump_packet))
// how many user bytes per file?
#define FILE_SIZE_BYTES  (RECORDS_PER_FILE*sizeof(struct dump_packet))
// how many blocks per file? Add 4k for directory structure and rounding.
#define FILE_SIZE_BLOCKS ((RECORDS_PER_FILE*sizeof(struct dump_packet))/1024 + 4)
// How many files to put in each directory.
#define DUMP_FILES_PER_DIR 100


// ------------------------------------------------------------
// These hosts are used by the 'killer' program and the kernel
// code. Update to reflect the hosts on your network that will
// be referred to by the 'killer' program or the kernel code.
//
// This static mapping is required as the kernel code is
// completely passive; it won't generate or respond to 'arp'
// requests.
// ------------------------------------------------------------
#define HOST_ADDRESSES_TABLE \
{ \
	{ "balance",     "192.168.3.1"    , "00:B0:D0:72:24:38" }, \
	{ "swift01",     "192.168.34.10"  , "00:90:27:E0:08:E1" }, \
	{ "swift02",     "192.168.34.20"  , "00:90:27:6A:F8:0A" }, \
	{ "swift09-eth0","192.168.34.90"  , "00:90:27:6A:F0:E8" }, \
	{ "swift09-eth1","10.0.3.1"       , "00:02:B3:9D:08:58" }, \
	{ "rapid5-eth0", "192.168.32.60"  , "00:A0:CC:50:E8:B0" }, \
	{ "rapid6-eth0", "10.0.2.7"       , "00:60:CF:20:51:A6" }, \
	{ "rapid6-eth1", "192.168.32.70"  , "00:60:CF:20:51:AD" }, \
	{ "rapid6-eth2", "10.0.1.7"       , "00:A0:CC:41:74:03" }, \
	{ "rapid7-eth0", "10.0.1.8"       , "00:A0:CC:40:BA:15" }, \
	{ "rapid7-eth1", "192.168.32.80"  , "00:60:CF:20:50:51" }, \
	{ "agile1-eth1", "192.168.33.20"  , "00:60:CF:20:51:B6" }, \
	{ "agile1-eth1", "192.168.33.20"  , "00:60:CF:20:51:B6" }, \
	{ "agile3-eth1", "192.168.33.40"  , "00:60:CF:20:50:97" }, \
    { "broadcast"  , "255.255.255.255", "FF:FF:FF:FF:FF:FF" }, /* special address, REQUIRED. */ \
    { NULL         ,  NULL            , "00:00:00:00:00:00" } /* terminator */ \
}


// ------------------------------------------------------------
// network port to put various things on. Currently only used:
//     in ../kernel/dump.c it sets port to which we send dump packets.
//     in ./user_network.c it sets port on which we listen.
//
// The kernel code, flooder, forger, and killer use this
// indirectly by referencing the two files listed above.
//
// How did I choose these numbers? Arbitrarily.
//     RADIANT = (18*1*4*9*1*14*20)%65536 = 50368
// Then I checked a very good ports list by neohapsis
//     http://www.neohapsis.com/neolabs/neo-ports/
// and ensured that nothing else commonly used this one.
// ------------------------------------------------------------
#define CLIENT_TICKET_PORT 50368
#define SERVER_TICKET_PORT 50369


// ------------------------------------------------------------
// Receive timeout, in seconds, for stuff done over UDP.
// Not really used at the moment.
// ------------------------------------------------------------
#define TICKET_RECV_TIMEOUT 5


// =============================================================================
// KERNEL LEVEL CONFIGURATION, interfaces...
// These had all better be in the "HOST_ADDRESSES_TABLE" above;
// if not you'll induce a kernel panic.
// =============================================================================

// Set up networking parameters.
// At 200ms and 1Gbps we will need to queue:
//     ((200/1000)(1000000000/8)/1500) =  16,666 packets.
//     At 1pkt/page we get:
//     16666*4096 =~ 65MB of memory required. Assume we have 128 for niceties.
// used in ../kernel/ticket_main.c
#define NETDEV_MAX_BACKLOG 20000 /* 20000 */
#define NO_CONG_THRESH      1000 /*  1000 */
#define NO_CONG             1000 /*  1000 */
#define LO_CONG            10000 /* 10000 */
#define MOD_CONG           18000 /* 18000 */

// The default input (collection) interface(s) on the collection box.
// Be sure to NULL terminate this list!!!
// TODO: Is this used correctly now??? No!
#define DUMP_INPUT_INTERFACES { "rapid4-eth1", "rapid4-eth2", NULL }

// The default output (dump) interface(s) on the collection box.
// Packets are sent using only the first output interface's source
// address, regardless of which interface they actually leave.
// (see dump_ap.src_hw)
// Be sure to NULL terminate this list!!!
// TODO: Is this used correctly now??? No!
#define DUMP_INPUT_INTERFACES { "rapid4-eth1", "rapid4-eth2", NULL }
#define DUMP_OUTPUT_INTERFACES { "rapid4-eth0", NULL }

// TODO To figure out which interfaces are *actually* used to output packets,
// TODO we stripe among [all interfaces]-[DUMP_INPUT]-[DEFAULT_DUMP_IGNORE].
// TODO was DUMP_IF_INPUT
// TODO was #define DUMP_IF_OUTPUT "rapid1"
// TODO #define DUMP_IF_IGNORE {NULL}

// The default input interface on the dump box, that is, the machine
// to which we shall address all our packets. (see dump_ap.dst_hw)

// TODO Note: If you're striping among multiple dump boxes, this breaks down;
// TODO I don't really want to deal with addressing each box independently...
// TODO Thus the use of 'broadcast' here.
// This list must be in sync with DUMP_OUTPUT_INTERFACES and NULL terminated!!!
#define DUMP_TARGET_INTERFACE { "broadcast", NULL }
// TODO #define DUMP_IF_TARGET "broadcast"

// Do we want to allow remote administration via raw ethernet packets?
// Define this if so.
//
// With this, we can reboot the machine or change states when routing.
//
// See netif_rx_* in ../kernel/network.c, ../kernel/dump.c, ../kernel/route.c
//     ../admin/admin.[ch]
#define ALLOW_INSECURE_ADMIN

// used in ../kernel/ticket_main.c
#define TICKET_MODE_DUMP  0
#define TICKET_MODE_ROUTE 1

// The mode in which to run the TICKET kernel.
// TODO: Make this dynamic (command line parse)!!!!
#define TICKET_MODE TICKET_MODE_ROUTE

// This is a table defining a state machine.
// Each line is a state (mode) with certain a certain delay & drop configuration.
//
//   idx     : Index, must be the actual C index into the array.
//   nextidx : Index of next state to which to move, when `duration' is over.
//   duration: How long to stay in this state:
//             3,600,000 is an hour (1000*60*60)
//             can't be longer than 2^32/1000 = 4,294,967ms = 4,294s = 71m.
//                 if you need it longer, put in duplicate states.
//   delay   : Delay in milliseconds before forwarding packets.
//   drop    : Drop every Nth packet (0=nothing, 1=everything)
//
// This is a SLOPPY delay mechanism; when the delay falls, any packets that are
//     queued use the new delay. That is, if several packets are queued with a
//     100ms delay and then the delay falls to 10ms, if packets are still in the
//     queue they will experience a delay between 10 and 100ms.
//
// Packet reordering *CANNOT* occur in the current implementation. So if you go
//     to a lower delay the first couple of packets will be delayed an arbitrary
//     value between the lower and higher delay.
//
// Watch your 'next_delay' indices, if they don't loop somewhere, you core dump.
//
// see ../kernel/route.c
//
// TODO: cache time of last administrative attempt; if get new one within a certain window, ignore it.
//       I put two identical entries here because I administer by broadcast;
//       two interfaces means two copies of the message means need two entries
//       for a single effect per administration attempt. (otherwise skip one).
#define ROUTING_CONFIG \
  /* idx, delay, duration, drop, nextidx */ \
	{  0,    50, 3600000,    0,       1 }, \
	{  1,    25, 3600000,    0,       2 }, \
	{  2,    25, 3600000,    0,       3 }, \
	{  3,    25, 3600000,    0,       4 }, \
	{  4,    50, 3600000,    0,       5 }, \
	{  5,    50, 3600000,    0,       6 }, \
	{  6,   100, 3600000,    0,       7 }, \
	{  7,   100, 3600000,    0,       0 }

// Interfaces between which to route packets when in routing mode.
// see ../kernel/route.c:configure_route_dev()
#define ROUTE_INTERFACE_0 "swift09-eth0"
#define ROUTE_INTERFACE_1 "swift09-eth1"

// _default_ verbosity: each -v flag on command line increments.
//    0: critical messages (errors) only.
//    1: status reports.
//    2: debugging information.
#define DEFAULT_VERBOSITY 1

#endif

