/*
###############################################################################
# This file is distributed as part of TICKET, but was written independently.
# It is Copyright (C) 2001 by Eric H. Weigle.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at weigle@cs.wisc.edu,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following website for updated information or other means by which he can
# be contacted:
# 	http://www.losalamos.com/ehw/
#
###############################################################################
*/
// #############################################################################
// Doubly linked list. Yes, I know...
//
// Deal with it, already.
// #############################################################################
#include <stdio.h>
#include <stdlib.h> // malloc, etc.
#include <errno.h>
#include <string.h> // strerror
#include "doubly.h"


#define BEFORE_ITEM 1
#define AFTER_ITEM 0

// =============================================================================
// Wrapper function for item creation.
// =============================================================================
static struct dlist_item *create_item(void *data) {
	struct dlist_item *tmp;

	// create item.
	if ((tmp=(struct dlist_item *)malloc(sizeof(struct dlist_item)))==NULL) {
		printf("Memory error: couldn't allocate new item!\n");
		exit(1);
	}
	tmp->data = data;

	return tmp;
}


// =============================================================================
// Generic function for removal of list items.
// =============================================================================
static void dlist_remove_item(struct dlinked_list *list, struct dlist_item *item) {
	// ------------------------------
	// handle forward pointers.
	// ------------------------------
	if (item->next) {
		// not tail
		item->next->prev = item->prev;
	} else {
		// tail
		list->tail = item->prev;

		if (list->tail) {
			// non-empty list
			list->tail->next = NULL;
		} else {
			// emtpy list
			list->head = NULL;
		}
	}

	// ------------------------------
	// handle backward pointers.
	// ------------------------------
	if (item->prev) {
		// not head
		item->prev->next = item->next;
	} else {
		// head
		list->head = item->next;

		if (list->head) {
			// non-empty list
			list->head->prev = NULL;
		} else {
			// emtpy list
			list->tail = NULL;
		}
	}

	// ------------------------------
	// done with memory.
	// ------------------------------
	free(item);
	list->count--;

	return;
}


// =============================================================================
// Generic function for insertion of list items
// =============================================================================
static void dlist_insert_item(struct dlinked_list *list, struct dlist_item *item, struct dlist_item *where, int before) {
	// error check.
	if (list==NULL) {
		printf("No list!\n");
		exit(1);
	}

	// special-case empty list.
	// where should only be NULL in this case.
	if (list->head==NULL) {
		list->head = list->tail = item;
		item->prev = item->next = NULL;
		list->count=1;
		return;
	}

	if (where==NULL) {
		printf("Bad call to dlist_insert_item!\n");
		exit(1);
	}

	if (before) {
		// insert before 'where'
		item->prev = where->prev;
		item->next = where;

		// special case head.
		if (where==list->head) {
			list->head = item;
		} else {
			where->prev->next = item;
		}

		where->prev = item;
	} else {
		// insert after 'where'
		item->next = where->next;
		item->prev = where;

		// special case tail.
		if (where==list->tail) {
			list->tail = item;
		} else {
			where->next->prev = item;
		}

		where->next = item;
	}

	list->count++;
}


// =============================================================================
// Find and remove element 'data' from this list.
// =============================================================================
int dlist_remove(struct dlinked_list *list, void *data) {
	struct dlist_item *curr;

	// error check.
	if (list==NULL) {
		printf("No list or head!\n");
		return 0;
	}

	// iterate through list.
	for (curr=list->head; curr!=NULL; curr=curr->next) {
		// try to find a match.
		if ((*list->compfn)(curr->data, data)==0) {
			dlist_remove_item(list, curr);

			return 1;
		}
	}

	// never found!
	return 0;
}


// =============================================================================
// Print the list to the given file.
// =============================================================================
void dlist_print(struct dlinked_list *list, FILE *file) {
	struct dlist_item *curr;

	// error check.
	if (list==NULL) {
		printf("No list or head!\n");
		return;
	}

	for (curr=list->head; curr; curr=curr->next) {
		fprintf(file, "(%p):\t", curr);
		(*list->printfn)(file, curr->data);
	}
	return;
}


// =============================================================================
// Remove and return the item at 'tail'.
// returns: 0 if list empty, 1 on success.
// =============================================================================
int dlist_pop_tail(struct dlinked_list *list, void **data) {
	// error check.
	if (list==NULL) {
		printf("No list!\n");
		exit(1);
	}

	if (list->tail==NULL) {
		// done poppin'; list empty.
		return 0;
	} else {
		// good data; pop.
		*data = list->tail->data;
		dlist_remove_item(list, list->tail);

		return 1;
	}
}

// =============================================================================
// Remove and return the item at 'head'.
// returns: 0 if list empty, 1 on success.
// =============================================================================
int dlist_pop_head(struct dlinked_list *list, void **data) {
	// error check.
	if (list==NULL) {
		printf("No list!\n");
		exit(1);
	}

	if (list->head==NULL) {
		// done poppin'; list empty.
		return 0;
	} else {
		// good item; pop.
		*data = list->head->data;
		dlist_remove_item(list, list->head);
		return 1;
	}
}


// =============================================================================
// Add an item to the list at its tail.
// =============================================================================
void dlist_push_tail(struct dlinked_list *list, void *data) {
	dlist_insert_item(list, create_item(data), list->tail, AFTER_ITEM);
}


// =============================================================================
// Add an item to the list at its head.
// =============================================================================
void dlist_push_head(struct dlinked_list *list, void *data) {
	dlist_insert_item(list, create_item(data), list->head, BEFORE_ITEM);
}


// =============================================================================
// Add an item to the list in sorted order.
// =============================================================================
void dlist_insert(struct dlinked_list *list, void *data) {
	struct dlist_item *curr;

	// error check.
	if (list==NULL || list->compfn==NULL) {
		printf("No list into which to insert, or compfn==NULL!\n");
		exit(1);
	}

	// find a place in the list.
	for (curr=list->head; curr!=NULL; curr=curr->next) {
		// check against current item.
		if (list->compfn(curr->data, data)>0) {
			dlist_insert_item(list, create_item(data), curr, BEFORE_ITEM);
			return;
		}
	}

	// special case tail (or empty list).
	dlist_insert_item(list, create_item(data), list->tail, AFTER_ITEM);

	return;
}


// =============================================================================
// Check that the list has valid structure.
// =============================================================================
int dlist_verify(struct dlinked_list *list) {
	struct dlist_item *item;
	int local_count;

	// stupid user error.
	if (list==NULL) {
		printf("Tried to validate a NULL list!\n");
		return 0;
	}

	// count items starting with head.
	local_count=0;
	for (item=list->head; item; item=item->next) {
		local_count++;
	}
	if (local_count!=list->count) {
		printf("From head, count %d != list count %d\n", local_count, list->count);
		return 0;
	}

	// count items starting with tail.
	local_count=0;
	for (item=list->tail; item; item=item->prev) {
		local_count++;
	}
	if (local_count!=list->count) {
		printf("From tail, count %d != list count %d\n", local_count, list->count);
		return 0;
	}

	// then head structure.
	if (list->head && list->head->prev!=NULL) {
		printf("head->prev!=NULL\n");
		return 0;
	}

	// then tail structure.
	if (list->tail && list->tail->next!=NULL) {
		printf("tail->next!=NULL\n");
		return 0;
	}

	return 1;
}

