/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Kernel ``dump'' mode, main file.
// functions for
//     packet allocation,
//     packet reception,
//     header collection,
//     packet transmission,
// etc.
// #############################################################################
#include "ticket.h"
#include "network.h"
#include "./common/shared_dump.h"
#include "./common/config.h"

#include <asm/atomic.h>

// dump_ap[0] is source, dump_ap[1] is dest, dump_ap[2] is NULL (// TODO? required???)
struct address_pair dump_ap[3];
char *collector_in[]=DUMP_INPUT_INTERFACES;
char *collector_out[]=DUMP_OUTPUT_INTERFACES;
char *dumper_in[]=DUMP_TARGET_INTERFACE;

// A semaphore for our init() to sleep on waiting for packets.
DECLARE_MUTEX_LOCKED(dump_init_sleep_sem);

// The list of packets we're working with
extern struct sk_buff_head ticket_skbs;


// =============================================================================
// Allocate a packet and reserve space for the headers.
// =============================================================================
static struct sk_buff *
alloc_packet(struct net_device *netdev) {
	struct sk_buff *skb;

	// ----------------------------------------
	// INITIALIZE SKB
	// ----------------------------------------
	// Allocate skb; We should have excess room to allow for alignment.
	// See header sizes in dump_common.h, it works out (?)
	if ((skb = alloc_skb(ETH_HEADER_LEN+ETH_DATA_LEN, GFP_ATOMIC))==NULL) {
		return NULL;
	}
	// fill in the skb with the parameters we need.
	skb->dev = netdev;

	// ----------------------------------------
	// RESERVE SPACE
	// ----------------------------------------
	skb_reserve(skb, ETH_HEADER_LEN+2); // reserve ethernet space (14+2=16, upper-aligned)
	skb_reserve(skb, IP_HEADER_LEN);    // reserve IP space, 20-byte IP header
	skb_reserve(skb, UDP_HEADER_LEN);   // reserve UDP space, 8 byte UDP header.
	skb_reserve(skb, DUMP_HEADER_LEN);  // reserve 8 bytes for sequence number of 'upper' header.

	// lay claim to the rest of the packet as our data area.
	(void)skb_put(skb, DUMP_RECORD_LEN*DUMPS_PER_PACKET);
	// zeroing it out _may_ not be required (???)
	memset(skb->data, 0, DUMP_RECORD_LEN*DUMPS_PER_PACKET);

	// Initialize our control buffer:
	//   skb->cb[0..7] for timestamp
	//      64 bit cycle counters
	//   skb->cb[8..11] for collection index
	//      (the place we're at in the packet (0..DUMPS_PER_PACKET-1))
	skb->cb[0] = skb->cb[1] = skb->cb[2]  = skb->cb[3]  = 0;
	skb->cb[4] = skb->cb[5] = skb->cb[6]  = skb->cb[7]  = 0;
	skb->cb[8] = skb->cb[9] = skb->cb[10] = skb->cb[11] = 0;

	// done.
	return skb;
}


// =============================================================================
// Fill in the headers, top down, that we reserved before.
// =============================================================================
static void
fill_headers(struct sk_buff *skb, unsigned long long seqno) {
	// ----------------------------------------
	// My simple header. Just the sequence no.
	// ----------------------------------------
	memcpy(skb_push(skb, DUMP_HEADER_LEN), &seqno, DUMP_HEADER_LEN);

	// ----------------------------------------
	// UDP header.
	// ----------------------------------------
	// UDP/TCP use parts of the IP header in their checksumming, hence the weird call series here.
	skb->h.uh = (struct udphdr *)skb_push(skb, UDP_HEADER_LEN);
	skb->h.uh->source = htons(CLIENT_TICKET_PORT);
	skb->h.uh->dest = htons(SERVER_TICKET_PORT);
	skb->h.uh->len = htons((DUMPS_PER_PACKET*DUMP_RECORD_LEN)+DUMP_HEADER_LEN+UDP_HEADER_LEN); // udp length includes header.

	// ARGH!!!!! Apparently, nobody really uses this checksum in ipv4...
	skb->h.uh->check=0;

	// ----------------------------------------
	// IP header.
	// ----------------------------------------
	skb->nh.iph = (struct iphdr *)skb_push(skb, IP_HEADER_LEN);
	skb->nh.iph->version = 4;
	skb->nh.iph->ihl = IP_HEADER_LEN>>2; // bytes --> 32-bit words
	skb->nh.iph->tos = 0; // change to one of the 'low latency' etc things?
	skb->nh.iph->tot_len = htons(ntohs(skb->h.uh->len)+IP_HEADER_LEN);
	skb->nh.iph->id = htons(seqno&0xffff); // kifes lowest 16 bits // TODO: May need to muck with this.
	skb->nh.iph->frag_off = htons(0);
	skb->nh.iph->ttl = 64; // Arbitrary.
	skb->nh.iph->protocol = IPPROTO_UDP;
	skb->nh.iph->saddr = dump_ap[0].ip_addr; // already in network byte order.
	skb->nh.iph->daddr = dump_ap[1].ip_addr; // already in network byte order.

	// By comparison, IP checksum is a cinch.
	skb->nh.iph->check = 0;
	skb->nh.iph->check = ip_fast_csum((unsigned char *)skb->nh.iph, skb->nh.iph->ihl);

	// ----------------------------------------
	// Ethernet header.
	// ----------------------------------------
	skb->mac.ethernet = (struct ethhdr *)skb_push(skb, ETH_HEADER_LEN);
	memcpy(skb->mac.ethernet->h_dest, dump_ap[1].mac_addr, skb->dev->addr_len);
	memcpy(skb->mac.ethernet->h_source, dump_ap[0].mac_addr, skb->dev->addr_len);
	skb->mac.ethernet->h_proto = htons(0x800); // IP TOS in NBO
}


// =============================================================================
// dump a binary buffer in hex.
// =============================================================================
void
bindump(void *buf, int len) {
	int i;
	unsigned char *ch=(unsigned char *)buf;

	for (i=0; i<len; i++) {
		printk("%02X ", ch[i]);
		if ((i+1)%16==0) {
			printk("\n");
		}
	}
	printk("\n");
}


// =============================================================================
// The actual function to collect information from the skbuff
// and save it to our headers structure.
// =============================================================================
void
collect_data(struct sk_buff *skb, struct collected_headers *hdr) {
	// first, copy length
	hdr->length = skb->len;

	// Then parse out information.
	if (skb->len >= ETH_HEADER_LEN) {
		// grab timestamp out of cb (control buffer)
		memcpy(&(hdr->timestamp), &(skb->cb[0]), 8);

		// Set up pointers
		skb->mac.ethernet = (struct ethhdr *)skb->data;

		// Grab ethernet header.
		memcpy(&(hdr->mac_dst[0]), &(skb->mac.ethernet->h_dest[0]), ETH_ALEN);
		memcpy(&(hdr->mac_src[0]), &(skb->mac.ethernet->h_source[0]), ETH_ALEN);
		hdr->toslen = skb->mac.ethernet->h_proto;

		// Allow remote killer code to reboot whilst in this mode.
		// TODO (?) remove for distribution?
		// TODO: make reference ALLOW_INSECURE_ADMIN (common/config.h)
		if (skb->len >= (ETH_HEADER_LEN+4)
			&& skb->mac.ethernet->h_proto==0x9999
			&& skb->data[14]=='D'
			&& skb->data[15]=='I'
			&& skb->data[16]=='E'
			&& skb->data[17]=='!') {
				ew_panic("Reboot request received!");
		}

		// Is it an IP packet? (0x800 NBO is 0x8 HBO due to endian flip!)
		if (hdr->toslen == 0x8 && skb->len >= IP_HEADER_LEN) {
			// Set up pointers
			skb_pull(skb, ETH_HEADER_LEN);
			skb->nh.iph = (struct iphdr *)(skb->data);

			// Grab IP header.
			hdr->ip_src     = skb->nh.iph->saddr;
			hdr->ip_dst     = skb->nh.iph->daddr;
			hdr->ip_len     = skb->nh.iph->tot_len;
			hdr->ip_hdr_len = (skb->nh.iph->ihl<<2); // do conversion words-->bytes
			hdr->protocol   = skb->nh.iph->protocol;

			// Grab TCP or UDP headers.
			// 20020912- oops, forgot to verify that it's not an IP frag!
			if (skb->len >= hdr->ip_hdr_len && skb->nh.iph->frag_off==0) {
				// Set up pointers
				skb_pull(skb, hdr->ip_hdr_len);

				// Is it a TCP or UDP packet?
				if (hdr->protocol==6 && skb->len >= TCP_HEADER_LEN) {
					// Set up pointers
					skb->h.th = (struct tcphdr *)(skb->data);
					// copy data
					hdr->src_port = skb->h.th->source;
					hdr->dst_port = skb->h.th->dest;
					hdr->upper_len.tcp_len.len = skb->h.th->doff; // `data offset'
					hdr->upper_len.tcp_len.pad = '\0'; // Now explicitly nulled (after first data set collected!)
				} else if (hdr->protocol==17 && skb->len >= UDP_HEADER_LEN) {
					// Set up pointers
					skb->h.uh = (struct udphdr *)(skb->data);
					// copy data
					hdr->src_port = skb->h.uh->source;
					hdr->dst_port = skb->h.uh->dest;
					hdr->upper_len.udp_len = skb->h.uh->len;
				}
			}
		}
	}
}


// =============================================================================
// The second-level 'main' function of the collection code.
//
// Called by ticket_init()
//
// This code is essentially the bottom half of the packet reception code:
//   netif_rx
//      get & queue packet, up()s semaphore this code sleeps upon.
//   The loop below then
//      allocates a collection packet (alloc_packet()),
//      collects headers into it (collect_data()),
//      fills in the destination headers (fill_headers()),
//      sends it off (hard_start_xmit()).
// =============================================================================
void
dump_init(char **args) {
	// The current packet to which we are dumping.
	//   Use skb->cb[0..7] for timestamp
	//          64 bit cycle counters
	//       skb->cb[8..11] for collection index
	//   	    (the place we're at in the packet (0..DUMPS_PER_PACKET-1))
	struct sk_buff *to_skb=NULL;
	// The current packet whose data is being collected.
	struct sk_buff *from_skb=NULL;
	// We use a sequence number to check reliability.
	unsigned long long seqno=0;
	// and a pointer to make lower code more readable.
	int *idx=NULL;
	// this is the next interface to send a dump packet over.
	int stripe_idx=0;
	// The actual number of ethernet devices
	int i, ethmax;

	// Then, see if we have parameters.
	//    <i=input(collection box input interface)>: for example, i=collector
	//    <o=output(collection box output interface)>: for example, o=collector
	//    <d=dest(dump box name)>: for example, d=dumper
	// TODO: THIS IS CURRENTLY REALLY BROKEN. FIX BEFORE DISTRIBUTION!!!!
	for (i=0; args[i]!=NULL; i++) {
		if (strncmp(args[i], "i=", 2)==0) {
			// Use user's collection box name
			collector_in[0]=&(args[i][2]);
		} else if (strncmp(args[i], "o=", 2)==0) {
			// Use user's collection box name
			collector_out[0]=&(args[i][2]);
		} else if (strncmp(args[i], "d=", 2)==0) {
			// Use user's dump box
			dumper_in[0]=&(args[i][2]);
		}
	}

	printk("Collector '%s' sending to dumper '%s'. Forging output as interface '%s'.\n", collector_in[0], dumper_in[0], collector_out[0]);

	// dump-specific configuration of devices.
	// first one is collector, rest are dumpers.
	if (!(ethmax=configure_dump_dev(dump_ap, collector_in, collector_out, dumper_in))) {
		ew_panic("Could not initialize dump devices!");
	}

	// Loop forever
	while (1) {
		// Put myself to sleep; when wakes, there's a packet to handle.
		down(&dump_init_sleep_sem);

		// Get the packet off the list.
		if ((from_skb = skb_dequeue(&ticket_skbs))==NULL) {
			printk("Dequeued NULL skb in init!!!\n");
			continue; // shouldn't happen, but don't die.
		}

		// Allocate a dump packet if required.
		if (to_skb==NULL) {
			if ((to_skb=alloc_packet(ew_eth[stripe_idx]))==NULL) {
				ew_panic("Failed to allocate a dump packet!");
			}
			// Stripe here; this code assumes the input
			//  interface(s) are the LAST ethernet interface(s).
			//  This order is enforced in network.c 'configure_dump_dev'
			stripe_idx = (stripe_idx+1)%ethmax;
		}

		// Get easier-to-use pointer of current dump header index.
		idx = ((int *)(&(to_skb->cb[8])));

		// ------------------------------------------------------------
		// Collect dump data into the my_index'th field of the skb.
		// ------------------------------------------------------------
		collect_data(from_skb, &(((struct collected_headers *)to_skb->data)[(*idx)++]));

		// -------------------------------------------------
		// IF packet is full, send it off.
		// -------------------------------------------------
		if (*idx >= DUMPS_PER_PACKET) {
			// fill in headers of packet
			fill_headers(to_skb, seqno++);

			// force the packet to be sent *NOW*
			//   the transmission code frees the skb.
			to_skb->dev->hard_start_xmit(to_skb, to_skb->dev);

			// This will make us allocate a new one on next loop.
			to_skb=NULL;
		}

		// --------------------------------------------------
		// Finish up, drop the packet we just processed.
		// --------------------------------------------------
		kfree_skb(from_skb);
	}
}


// =============================================================================
// the callback to replace netif_rx; does dumping.
// =============================================================================
int
netif_rx_dump(struct sk_buff *skb) {
	register unsigned int dummylo;
	register unsigned int dummyhi;

	// Get the timestamp into the cb area.
	// Chose to inline versus calling get_timestamp for efficiency.
	asm volatile (".byte 0x0f, 0xa2" : : : "eax", "ebx", "ecx", "edx");
	asm volatile (".byte 0x0f, 0x31;              # the RDTSC instruction
					movl %%eax, %0
					movl %%edx, %1"
					: "=r" (dummylo), "=r" (dummyhi)
					:
					: "eax", "edx"
					);
	*((unsigned long long *)skb->cb) = ((unsigned long long)dummyhi)<<32 | dummylo;

	// First, check the packet. It should have the eth. headers removed,
	// but still valid (in the packet)
	if ((skb->data - skb->mac.raw)==ETH_HEADER_LEN) {
		// We want the headers back so we can look at 'em.
		skb_push(skb, ETH_HEADER_LEN);

#ifdef ALLOW_INSECURE_ADMIN
		// (Network reboot code. C break-out logic makes it fairly low cost.)
		if (skb->len >= (ETH_HEADER_LEN+4)
			&& skb->mac.ethernet->h_proto==0x9999
			&& skb->data[14]=='D'
			&& skb->data[15]=='I'
			&& skb->data[16]=='E'
			&& skb->data[17]=='!') {
				ew_panic("Reboot request received!");
		}
#endif

		// Enqueue the packet for later.
		// Gets spinlock, so is 'safe'
		skb_queue_tail(&ticket_skbs, skb);

		// notify init that there's a packet.
		up(&dump_init_sleep_sem);

		return NET_RX_CN_LOW; // claim low congestion...
	}

	// Packet failed ethernet check! Just destroy it.
	// make sure to note this drop, though.
	printk("Got a freaky packet!\n");
	bindump(skb->data, skb->len);
	kfree_skb(skb);
	ticket_drops++;

	return NET_RX_DROP;
}

