/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Kernel networking functions.
// #############################################################################
#include "ticket.h"
#include "network.h"
//  This is pretty wack, but I want to include just this data in my userspace
//  killer program too, and this is the easiest way (#include)
#include "./common/shared_addresses.h"
#include "./common/config.h"

// for rtnl_shlock(), etc (need to hold to use dev_get_by_index, dev_open)
#include <linux/rtnetlink.h>


// =============================================================================
// File-global variables.
// =============================================================================
// name mapping; passive so don't do ARP stuff.
struct name_mapping_raw map_addrs_raw[] = HOST_ADDRESSES_TABLE;
struct name_mapping     *map_addrs;

// 'global' pointers to network devices.
struct net_device *ew_eth[ETHERNET_IF_COUNT] = ETHERNET_IF_INITIALIZER;

// How many packets dropped so far?
unsigned long ticket_drops=0;

// global pointer to the receive function, too
int (*netif_rx)(struct sk_buff *skb)=&netif_rx_drop;


// =============================================================================
// print out all interfaces in the table. Pass in a message to
// print, and the number of input interfaces.
// =============================================================================
void print_interfaces(char *msg, int num_input) {
	struct name_mapping *addrptr;
	int idx;
	char tmpbuf1[32];
	char tmpbuf2[32];

	printk(KERN_WARNING "All ew_eth devices (%s):\n", msg);
	// Print out the link information.
	for (idx=0; ew_eth[idx]!=NULL; idx++) {
		if (!(addrptr = get_addr_info(map_addrs, NULL, 0, ew_eth[idx]->dev_addr))) {
			char tmpbuf[32];
			printk(KERN_EMERG "Failed match of device in this machine (ew_eth[%d], '%s'):\n", idx, ew_eth[idx]->name);
			printk(KERN_EMERG "Address: %s\n", mac2str(ew_eth[idx]->dev_addr, tmpbuf));
			printk(KERN_EMERG "Table searched:\n");
			print_mapping(map_addrs);
			ew_panic("Failed get address information!");
		}

		printk(KERN_WARNING "    ew_eth[%d]=\"%s\", \"%s\", \"%s\", \"%s\" (%s interface)\n",
				idx, ew_eth[idx]->name, addrptr->name,
				mac2str(addrptr->mac, tmpbuf1),
				ip2str_nbo(addrptr->ip, tmpbuf2),
				(idx<num_input ? "an output": "an input")
				);
	}
}


// =============================================================================
// Set up this structure for source and destination address.
// =============================================================================
void fill_addr_pairs(struct address_pair *ap, char *src, char *dst) {
	struct name_mapping *addrptr;

	// find the mac address given the string...
	if (!(addrptr = get_addr_info(map_addrs, src, 0, NULL))) {
		ew_panic("Failed fill_addr_pairs (src)!");
	}
	memcpy(ap[0].mac_addr, addrptr->mac, ETH_ALEN);
	ap[0].ip_addr = addrptr->ip;

	if (!(addrptr = get_addr_info(map_addrs, dst, 0, NULL))) {
		ew_panic("Failed fill_addr_pairs (dst)!");
	}
	memcpy(ap[1].mac_addr, addrptr->mac, ETH_ALEN);
	ap[1].ip_addr = addrptr->ip;

	// null-terminate
	memset(&(ap[2]), 0, sizeof(struct address_pair));
}


// =============================================================================
// Bring up network interfaces.
//
// We do here in the kernel what a
//    '/sbin/ifconfig eth0 up 192.168.xx.xx netmask 0xffff0000'
// does in userland.
//
//  Find all ethernet devices in system and get pointers;
//      bring up these interfaces.
//  Sort our pointers to them based on MAC address
//    critical--> Uniform order in my reference array,
//                regardless of probe responses.
// =============================================================================
int ifupall() {
	struct net_device *netdev;
	int devnum, ethnum, i, j;

	// initialize pointers
	for (ethnum=0; ethnum<ETHERNET_IF_COUNT; ethnum++) {
		ew_eth[ethnum] = NULL;
	}

	// set up address mapping.
	map_addrs = process_mapping(map_addrs_raw);

	// ----------------------------------------
	// Find and open ethernet devices
	// ----------------------------------------
	// * We know that the function net_device_init in drivers/net/setup.c has executed.
	//    (via net_dev_init in net/core/dev.c);
	//    Thus, the drivers *should* be properly configured by this point.

	rtnl_shlock(); // need to hold to use dev_get_by_index, dev_open.
	// Loop through indices to find our ethernet cards.
	// devnum 0 is invalid for some reason, start with 1.
	for (devnum=1, ethnum=0; (netdev=dev_get_by_index(devnum))!=NULL && ethnum<ETHERNET_IF_COUNT; devnum++) {
		// netdev is a valid device; what the heck is it?
		//    see include/linux/netdevice.h

		// Is it an ethernet interface?
		// Ok, can do this more efficiently with device type field, whatever.
		if (strncmp(netdev->name, "eth", 3)==0) {
			// OK, save pointer to device for later
			ew_eth[ethnum++] = netdev;

			// Finally, open it up (the actual 'up' routine)
			if (dev_open(netdev)<0) {
				printk("Failed to bring up interface %s!!!\n", netdev->name);
				return 0;
			}
		} else {
			// give back the interface. Decrements use count.
			dev_put(netdev);
		}
	}
	rtnl_shunlock();

	// error check.
	if (ethnum<1) {
		printk("Too few ethenet devices in system!\n");
		return 0;
	}

	// ------------------------------------------------------------
	// sort devices by name, using stupid O(N*N) swapping sort.
	// Ordering defined by PROBE order, and cannot be assumed.
	// ------------------------------------------------------------
	for (i=0; i<(ethnum-1); i++) {
		for (j=i+1; j<ethnum; j++) {
			if (strncmp(ew_eth[i]->name, ew_eth[j]->name, 4)>0) {
				// flip them, in wrong order.
				netdev = ew_eth[i];
				ew_eth[i] = ew_eth[j];
				ew_eth[j] = netdev;
			}
		}
	}

	// Print out the link information.
	print_interfaces("end of ifupall()", 0);

	// Wait 10 seconds for interfaces to come up.
	printk("Waiting for ifup...");
	for (i=10; i>0; i--) {
		printk("(%d)", i);
		mdelay(1000);
	}
	printk("Ok.\n");

	// done!
	return 1;
}


// =============================================================================
// find MAC in the device list (ew_eth), and move to position 'headnum'
// =============================================================================
int move_to_head(int head_num, char *mac) {
	struct net_device *netdev;
	int dev_num;

	for (dev_num=head_num; ew_eth[dev_num]!=NULL; dev_num++) {
		if (memcmp(ew_eth[dev_num]->dev_addr, mac, ETH_ALEN)==0) {
			netdev = ew_eth[dev_num];
			ew_eth[dev_num] = ew_eth[head_num];
			ew_eth[head_num] = netdev;
			return 1;
		}
	}
	return 0;
}


// =============================================================================
// Configure the routing device
// TODO don't limit to two devices
// =============================================================================
void configure_route_dev(char *if0, char *if1) {
	struct name_mapping *addrptr;
	struct net_device *netdev;
	int i;

	// Mapping:
	//         ew_eth[0] = if0
	//         ew_eth[1] = if1
	// Forward:
	//         ew_eth[0] --> ew_eth[1], rewrite source as ew_eth[1], rewrite target as ap->dst_hw
	//         ew_eth[1] --> ew_eth[0], rewrite source as ew_eth[0], rewrite target as ap->src_hw

	// ------------------------------------------------------------
	// first, get interface info set up into ew_eth[0] and ew_eth[1].
	// ------------------------------------------------------------

	// --------------------
	// --> ew_eth[0]
	// --------------------
	if (!(addrptr = get_addr_info(map_addrs, if0, 0, NULL))) {
		printk(KERN_EMERG "Failed to match interface '%s'\n", if0);
		ew_panic("Failed matching first routing device!");
	}
	// Then put it FIRST
	for (i=0; ew_eth[i]!=NULL; i++) {
		if (memcmp(ew_eth[i]->dev_addr, addrptr->mac, ETH_ALEN)==0) {
			netdev = ew_eth[i];
			ew_eth[i] = ew_eth[0];
			ew_eth[0] = netdev;
			break;
		}
	}
	if (ew_eth[i]==NULL) {
		ew_panic("Could not find interface 0 in this box!\n");
	}

	// --------------------
	// --> ew_eth[1]
	// --------------------
	if (!(addrptr = get_addr_info(map_addrs, if1, 0, NULL))) {
		printk(KERN_EMERG "Failed to match interface '%s'\n", if1);
		ew_panic("Failed matching second routing device!");
	}
	// Then put it SECOND
	for (i=1; ew_eth[i]!=NULL; i++) {
		if (memcmp(ew_eth[i]->dev_addr, addrptr->mac, ETH_ALEN)==0) {
			netdev = ew_eth[i];
			ew_eth[i] = ew_eth[1];
			ew_eth[1] = netdev;
			break;
		}
	}
	if (ew_eth[i]==NULL) {
		ew_panic("Could not find interface 1 in this box!\n");
	}

	// --------------------
	// kill any superfluous interfaces.
	// --------------------
	for (i=2; ew_eth[i]!=NULL; i++) {
		// otherwise, remove the device from our list.
		rtnl_shlock(); // need to hold to use dev_get_by_index, dev_open.
			// close device.
			dev_close(ew_eth[i]);
			// give back the interface. Decrements use count.
			dev_put(ew_eth[i]);
		rtnl_shunlock();
		ew_eth[i]=NULL;
	}

	// wise man say: transparent router must hear all and say little.
	dev_set_promiscuity(ew_eth[0], 1);
	dev_set_promiscuity(ew_eth[1], 1);

	print_interfaces("end of configure_route_dev()", 0); // TODO prints input/output type wrong, but who cares.
}


// =============================================================================
// Set appropriate promiscuity and addresses for later.
// =============================================================================
int configure_dump_dev(struct address_pair *ap, char **collector_in, char **collector_out, char **dumper) {
	struct name_mapping *addrptr;
	int collector_num, min_dev_num, dev_num, retval;

	// Configure addresses for the 'forged' packets we'll send later.
	// Note that we do NOT use the dev->dev_addr because sometimes we WANT to forge the MAC addresses.
	// TODO this doesn't currently handle multiple output interfaces!
	fill_addr_pairs(ap, collector_out[0], dumper[0]);

	// ------------------------------------------------------------
	// Get list of devices, such that:
	//    all output devices are in the list and first, AND
	//    all input devices are in the list and last, AND
	//    no other devices are in the list.
	// ------------------------------------------------------------
	min_dev_num=0;

	// ------------------------------
	// Put output devices FIRST.
	// ------------------------------
	for (collector_num=0; collector_out[collector_num]!=NULL; collector_num++) {
		if (!(addrptr = get_addr_info(map_addrs, collector_out[collector_num], 0, NULL))) {
			ew_panic("Failed match of dump output device");
		}

		// find in list, move to head of list.
		if (move_to_head(min_dev_num, addrptr->mac)) {
			min_dev_num++;
		} else {
			char tmpbuf[32];
			// no match!
			printk(KERN_EMERG "Could not match input device number %d, %s \n", collector_num, mac2str(addrptr->mac, tmpbuf));
			return 0;
		}
	}
	// later on we'll return this so we can do a simple modulo over the
	// output devices for round-robin striping of output.
	retval = min_dev_num;

	// ------------------------------
	// Put input devices next; will be LAST after removal of extraneous devices.
	// ------------------------------
	for (collector_num=0; collector_in[collector_num]!=NULL; collector_num++) {
		if (!(addrptr = get_addr_info(map_addrs, collector_in[collector_num], 0, NULL))) {
			ew_panic("Failed match of dump input device");
		}

		// find in list, move to head of list.
		if (move_to_head(min_dev_num, addrptr->mac)) {
			// put collection device into promiscuous mode.
			dev_set_promiscuity(ew_eth[min_dev_num], 1);
			min_dev_num++;
		} else {
			char tmpbuf[32];
			// no match!
			printk(KERN_EMERG "Could not match output device number %d, %s \n", collector_num, mac2str(addrptr->mac, tmpbuf));
			return 0;
		}
	}

	// ------------------------------
	// all remaining devices will be ignored.
	// ------------------------------
	for (dev_num=min_dev_num; ew_eth[dev_num]!=NULL; dev_num++) {
		rtnl_shlock(); // need to hold to use dev_get_by_index, dev_open.
			// close device.
			dev_close(ew_eth[dev_num]);
			// give back the interface. Decrements use count.
			dev_put(ew_eth[dev_num]);
		rtnl_shunlock();

		ew_eth[dev_num] = NULL;
	}

	print_interfaces("end of configure_dump_dev()", retval);

	// Woo-hoo, all interfaces are up and running.
	return (retval);
}


// =============================================================================
// Handle reception of a packet for modes where just want to drop,
// such as before I'm fully initialized.
// =============================================================================
int netif_rx_drop(struct sk_buff *skb) {
	// This code allows network reboots while in this mode.
	if (skb==NULL || skb->dev==NULL ||
		(skb->data - skb->mac.raw)!=ETH_HEADER_LEN)
		goto out;

	skb_push(skb, ETH_HEADER_LEN);
	if (skb->len < ETH_HEADER_LEN)
		goto out;

#ifdef ALLOW_INSECURE_ADMIN
	// (Network reboot code. C break-out logic makes it fairly low cost.)
	if (skb->len >= (ETH_HEADER_LEN+4) //+IP_HEADER_LENGTH=20
		&& skb->mac.ethernet->h_proto==0x9999
		&& skb->data[14]=='D'
		&& skb->data[15]=='I'
		&& skb->data[16]=='E'
		&& skb->data[17]=='!') {
			ew_panic("Reboot request received!");
	}
#endif

out:
	kfree_skb(skb);
	ticket_drops++;
	return NET_RX_DROP;
}

