/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Routing mode. use TICKET as a dumb WAN emulator.
// #############################################################################
#include "ticket.h"
#include "network.h"
#include "./common/config.h"
#include "./common/shared_addresses.h"

// from network.c
extern struct name_mapping *map_addrs;

// configuration. see common/config.h for more on this structure.
struct routing_config {
	unsigned long idx;      // UNUSED. Just for convenience in writing.
	unsigned long delay;    // milliseconds to delay each packet.
	unsigned long duration; // how long (milleseconds) to maintain this delay mode. (SEMANTICS CHANGE WARNING: 0 means core dump!)
	unsigned long drop_n;   // drop every Nth packet. (0 means drop nothing. 1 means drop everything)
	int next_delay;         // 'goto' index in array of next delay when this one expires.
};

// =============================================================================
// File-global variables.
// =============================================================================
// config parameters.
extern int netdev_max_backlog;
extern int no_cong_thresh;
extern int no_cong;
extern int lo_cong;
extern int mod_cong;

// The delay characteristics we have. See ./common/config.h
struct routing_config routing_states[] = { ROUTING_CONFIG };

// The queue to delay skbs.
extern struct sk_buff_head ticket_skbs;

// A semaphore for our local init() to sleep on...
DECLARE_MUTEX_LOCKED(route_init_sleep_sem);


// =============================================================================
// Route a packet (socket buffer).
//   Remember can only use spinlocks!
// =============================================================================
int netif_rx_route(struct sk_buff *skb) {
	// First, check the packet. It should have the eth. headers removed,
	// but still valid (in the packet).
	if (skb==NULL || skb->dev==NULL ||
		(skb->data - skb->mac.raw)!=ETH_HEADER_LEN) {
		ew_panic("Nonforwardable packet recieved.");
	}

	// can only queue so many packets before we're screwed.
	if (ticket_skbs.qlen>=netdev_max_backlog) {
		kfree_skb(skb);
		ticket_drops++;
		return NET_RX_DROP;
	}

	// Enqueue the packet for later.
	// Gets spinlock, so is 'safe'
	skb_queue_tail(&ticket_skbs, skb);

	// timestamp; after queueing
	// (trying to minimize work done before queueing;
	//    queue length value checks first!)
 	if (skb->stamp.tv_sec == 0) {
 		// 20021205; this got renamed between 2.4.5 and 2.4.20... get_fast_time --> do_gettimeofday
 		do_gettimeofday(&skb->stamp);
	}

	// notify init that there's a packet.
	up(&route_init_sleep_sem);

	// return congestion rating;
	// shortest path occurs with highest congestion.
	if (ticket_skbs.qlen>=mod_cong) {
		return NET_RX_CN_HIGH;
	} else if (ticket_skbs.qlen>=lo_cong) {
		return NET_RX_CN_MOD;
	} else if (ticket_skbs.qlen>=no_cong) {
		return NET_RX_CN_LOW;
	} else {
		return NET_RX_SUCCESS;
	}
}


// =============================================================================
// get diff in microseconds.
// `later' damn well be after `before'.
// =============================================================================
unsigned long mu_s_diff(struct timeval *before, struct timeval *later) {
	unsigned long diff;

	diff = (1000000 * ((unsigned long)later->tv_sec - (unsigned long)before->tv_sec));
	diff += (unsigned long)later->tv_usec;
	diff -= (unsigned long)before->tv_usec;

	return(diff);
}


// =============================================================================
// return next delay index, tell user what's up.
// TODO: proceed past other states by looping and time subtraction;
//       may need to move past multiple states at once!
// =============================================================================
int change_delay(int cur_delay, struct timeval *tv_start, struct timeval *tv_now, int *drop_idx) {
	// move to next state.
	cur_delay=routing_states[cur_delay].next_delay;

	printk("Delay change. state[%d]: %lu ms, Duration: %ld ms, Drop every %ldth, Next index %d.\n",
			cur_delay,
			routing_states[cur_delay].delay,
			routing_states[cur_delay].duration,
			routing_states[cur_delay].drop_n,
			routing_states[cur_delay].next_delay);

	// book-keeping.
	*drop_idx=0;
	do_gettimeofday(tv_start); // 20021205; Between 2.4.5 & 2.4.20... 'get_fast_time' renamed to 'do_gettimeofday'
	do_gettimeofday(tv_now);   // required to avoid double progression

	// Done! return new delay index.
	return(new_idx);
}


// =============================================================================
// 'main' function for this code.
// =============================================================================
void route_init(char **args) {
	struct sk_buff *skb;
	struct timeval tv_start, tv_now;
	int drop_idx;
	int cur_delay=0;
	struct name_mapping *dst_addr, *broadcast_addr;
	// debug:	char tmpbuf1[32], tmpbuf2[32];

	// get pointer to broadcast information for later checks.
	if (!(broadcast_addr=get_addr_info(map_addrs, "broadcast", 0, NULL))) {
		ew_panic("Must have broadcast address in shared_addresses configuration table!!!!");
	}

	// convert delay from ms to us; do this before the print to detect funkiness.
 	// 20021205; this got renamed between 2.4.5 and 2.4.20... get_fast_time --> do_gettimeofday
	do_gettimeofday(&tv_start);
	printk("%d: Delaying packets for %lu ms each way, duration %ld.\n", cur_delay, routing_states[cur_delay].delay, routing_states[cur_delay].duration);

	// TODO: needs to be command line parameterized. See code in dump.c
	configure_route_dev(ROUTE_INTERFACE_0, ROUTE_INTERFACE_1);

	drop_idx=0;
	while (1) {
		// wait for a packet.
		down(&route_init_sleep_sem);

		// remove packet from head of queue
		skb = skb_dequeue(&ticket_skbs);
		drop_idx++;
		// 20021205; this got renamed between 2.4.5 and 2.4.20... get_fast_time --> do_gettimeofday
		do_gettimeofday(&tv_now);

		// We want the headers back!
		skb->nh.iph = (struct iphdr *)skb->data;
		skb_push(skb, ETH_HEADER_LEN);
		if (skb->len < ETH_HEADER_LEN) {
			ew_panic("Got a freakishly short ethernet packet!");
		}
		skb->mac.ethernet = (struct ethhdr *)skb->data;

#ifdef ALLOW_INSECURE_ADMIN
		// TODO Add other runtime configuration options (see admin directory)
		// TODO Change to be common for administration
		//
		// (Network reboot code. C break-out logic makes it fairly low cost.)
		if (skb->mac.ethernet->h_proto==0x9999 && skb->len >= (ETH_HEADER_LEN+4)) {
			if (memcmp(&(skb->data[ETH_HEADER_LEN]), "DIE!", 4)==0) {
				ew_panic("Reboot request received!");
			} else if (memcmp(&(skb->data[ETH_HEADER_LEN]), "NEXT", 4)==0) {
				cur_delay = change_delay(cur_delay, &tv_start, &tv_now, &drop_idx);
			}
			goto drop;
		}
#endif

		// ----------------------------------------
		// Packets to drop
		// ----------------------------------------
		// debug: printk("Ethernet: Drop broadcast packet (destination already should get) from %s\n", mac2str(skb->mac.ethernet->h_source, tmpbuf1));
		if (!memcmp(skb->mac.ethernet->h_dest, broadcast_addr->mac, ETH_ALEN)) { goto drop; }

		// debug: printk("Ethernet: Drop looped packet (source MAC is that of device 0)\n");
		if (memcmp(ew_eth[0]->dev_addr, skb->mac.ethernet->h_source, ETH_ALEN)==0) { goto drop; }

		// debug: printk("Ethernet: Drop looped packet (source MAC is that of device 1)\n");
		if (memcmp(ew_eth[1]->dev_addr, skb->mac.ethernet->h_source, ETH_ALEN)==0) { goto drop; }

		// debug: printk("Ethernet: Drop non-IP packet (can't route), protocol 0x%04X\n", ntohs(skb->mac.ethernet->h_proto));
		if (skb->mac.ethernet->h_proto!=htons(ETH_P_IP) || skb->len < (ETH_HEADER_LEN+IP_HEADER_LEN)) { goto drop; }

		// debug: printk("IP: Drop broadcast packet (destination already should get) from %s\n", ip2str_nbo(skb->nh.iph->saddr, tmpbuf1));
		if (skb->nh.iph->daddr==broadcast_addr->ip) { goto drop; }

		// debug: printk("IP: Drop un-routable packet (failed to match IP address %s)\n", ip2str_nbo(skb->nh.iph->daddr, tmpbuf1));
		if (!(dst_addr=get_addr_info(map_addrs, NULL, skb->nh.iph->daddr, NULL))) { goto drop; }

		// debug: printk("Drop packets if user says so in routing table... it's time!\n");
		if (routing_states[cur_delay].drop_n && drop_idx==routing_states[cur_delay].drop_n) { goto drop; }

		// ----------------------------------------
		// Has packet been delayed long enough? (remember, routing_states in MILLISECONDS, timevals in MICROSECONDS)
		// ----------------------------------------
		while (mu_s_diff(&skb->stamp, &tv_now)<(routing_states[cur_delay].delay*1000)) {
			// this gives a 1ms granularity... good enough?
			// might want to use udelay instead...
			mdelay(1);
			// 20021205; this got renamed between 2.4.5 and 2.4.20... get_fast_time --> do_gettimeofday
			do_gettimeofday(&tv_now);
		}

		// Which interface to output packet on (Opposite than received)
		if ((void *)skb->dev == (void *)ew_eth[0]) {
			skb->dev = ew_eth[1];
		} else if ((void *)skb->dev == (void *)ew_eth[1]) {
			skb->dev = ew_eth[0];
		} else {
			// debug: printk("skb received on bad interface (%s), being ignored.\n", skb->dev->name);
			goto drop;
		}

		// Put in proper MAC addresses...
		//   source address should become:      OUTGOING INTERFACE
		//   destination address should become: TARGET BOX GIG-E INTERFACE
		memcpy(skb->mac.ethernet->h_source, skb->dev->dev_addr, skb->dev->addr_len);
		memcpy(skb->mac.ethernet->h_dest,   dst_addr->mac,      skb->dev->addr_len);

		// force the packet to be sent *NOW*. The transmission code frees the skb.
		// Don't muck with checksum... Otherwise will re-checksum with wrong pointers and corrupt packet!!!!
		// debug: printk("Sending to %s, %s, %s via %s.\n", dst_addr->name, ip2str_nbo(skb->nh.iph->daddr, tmpbuf1), mac2str(skb->mac.ethernet->h_dest, tmpbuf2), skb->dev->name);
		skb->ip_summed = CHECKSUM_NONE;
		// 20020912: spin lock should not be required; there is no contention.
		// spin_lock(&skb->dev->xmit_lock);
		skb->dev->hard_start_xmit(skb, skb->dev);
		// spin_unlock(&skb->dev->xmit_lock);

loop:
		// handle delay characteristics.
		if (mu_s_diff(&tv_start, &tv_now)>= (1000*routing_states[cur_delay].duration)) {
			cur_delay = change_delay(cur_delay, &tv_start, &tv_now, &drop_idx);
		}
	}
drop:
	drop_idx=0;
	kfree_skb(skb);
	ticket_drops++;
	goto loop;
}

