/*
###############################################################################
# This file is distributed as part of TICKET, but was written independently.
# It is Copyright (C) 2001 by Eric H. Weigle.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at weigle@cs.wisc.edu,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following website for updated information or other means by which he can
# be contacted:
# 	http://www.losalamos.com/ehw/
#
###############################################################################
*/
// #############################################################################
// Basic linked hash table code, code itself.
// #############################################################################
#include <stdio.h>
#include <stdlib.h> // malloc, etc.
#include <errno.h>
#include <string.h> // strerror
#include "hashtable.h"

// =============================================================================
// A ``new'' wrapper that'll die on errors.
// =============================================================================
static void *local_new(unsigned long size) {
	void *data;

	if (size==0 || (data=calloc(1, size))==NULL) {
		printf("Failed to allocate %lu bytes!\n", size);
		exit(1);
	}

	return data;
}


// =============================================================================
// A ``free'' wrapper. Useful for debugging.
// =============================================================================
static void local_destroy(void *data) {
	// printf("Freeing data at %08X\n", data);
	free(data);
}


// =============================================================================
// Create a new hashtable with given hash functions.
// =============================================================================
int hash_create(struct hashtable *table, int size, unsigned int (*hashfn)(void *, int), int (*compfn)(const void *, const void *), int (*printfn)(FILE *file, const void *)) {
	// error check
	if (table==NULL || size<=0 || hashfn==NULL || compfn==NULL) {
		printf("Failed to find intelligent user!\n");
		return -1;
	}

	// get memory for table (array of pointers to hashtable_items)
	table->table = (struct hashtable_item **)local_new(size*sizeof(struct hashtable_item *));

	// initialize.
	table->nitems = 0;
	table->size = size;
	table->hashfn = hashfn;
	table->compfn = compfn;
	table->printfn = printfn;
	return 1;
}


// =============================================================================
// Destroy the hash table.
// =============================================================================
int hash_destroy(struct hashtable *table) {
	// error check
	if (table==NULL || table->nitems>0) {
		printf("Bad table; null or non-empty. Use hash_remove or hash_get_next!!!\n");
		return -1;
	}

	// actually release up the memory we've taken
	local_destroy(table->table);
	return 1;
}


// =============================================================================
// call in loop (while (get_next(...); free(item))) before destroying hash
// If count is not null, returns count in the hash item (number of collisions)
// returns -1 on failure, index on success
// =============================================================================
static int hash_get_next(struct hashtable *table, void **item, int *idx, int *magic) {
	struct hashtable_item *item_hash;

	// error check
	if (table==NULL || table->nitems<=0 || item==NULL || idx==NULL || *idx<0 || *idx>=table->size) {
		printf("Stupid user error in hash_get_next (bad parameter)!\n");
		return -1;
	}

	// loop through table until you find a hash item.
	for ( ; *idx<table->size; (*idx)++) {
		if (table->table[*idx] != NULL) {
			item_hash = table->table[*idx];
			table->table[*idx] = item_hash->next;
			*item = item_hash->item;
			if (magic!=NULL) {
				*magic = item_hash->magic;
			}
			local_destroy(item_hash);
			table->nitems--;
			return *idx;
		}
	}

	printf("Failed to find an item at or greater than index %d!\n", *idx);
	return -1;
}


// =============================================================================
// remove the given item from the table, if it exists.
// =============================================================================
int hash_remove(struct hashtable *table, void *item, int *offset, int *magic) {
	int idx, jdx;
	struct hashtable_item *item_hash, *item_last;

	// error check
	if (table==NULL || table->nitems<=0) {
		printf("Table is empty or item is NULL (hash_remove)!\n");
		return -1;
	}

	// hash.
	idx = table->hashfn(item, table->size);
	if (table->table[idx]==NULL) {
		// printf("Not found in table!\n");
		return -1;
	}

	// find and remove from list.
	for (jdx=0, item_hash=item_last=table->table[idx]; item_hash; item_hash=item_hash->next, jdx++) {
		if (table->compfn(item_hash->item, item)==0) {
			// Found item. Get requested data.
			if (offset!=NULL) {
				*offset = jdx;
			}
			if (magic!=NULL) {
				*magic = item_hash->magic;
			}

			// Remove from list.
			if (item_hash==item_last) {
				// first item
				table->table[idx] = item_hash->next;
			} else {
				// item elsewhere in list
				item_last = item_hash->next;
			}
			table->nitems--;

			// clean up and leave.
			local_destroy(item_hash);
			return idx;
		}
		item_last = item_hash;
	}

	// not found in list, either.
	return -1;
}


// =============================================================================
// returns the item index from the table, if it exists, WITHOUT REMOVING IT.
// =============================================================================
int hash_find(struct hashtable *table, void *item, int *offset, int *magic) {
	int idx, jdx;
	struct hashtable_item *item_hash;

	// error check
	if (table==NULL || table->nitems<=0) {
		printf("Table is empty (hash_find)!\n");
		return -1;
	}

	// hash.
	idx = table->hashfn(item, table->size);
	if (table->table[idx]==NULL) {
		// printf("Not found in table!\n");
		return -1;
	}

	// is it in the list at this index?
	for (jdx=0, item_hash=table->table[idx]; item_hash; item_hash=item_hash->next, jdx++) {
		if (table->compfn(item_hash->item, item)==0) {
			if (offset!=NULL) {
				*offset = jdx;
			}
			if (magic!=NULL) {
				*magic = item_hash->magic;
			}
			return idx;
		}
	}

	// not found in list, either.
	return -1;
}


// =============================================================================
// add the given item to the table, if it exists.
// =============================================================================
unsigned int hash_insert(struct hashtable *table, void *item, unsigned int magic) {
	int idx;
	struct hashtable_item *item_hash, *table_hash;

	// hash.
	idx = table->hashfn(item, table->size);

	// check for collisions with possibly existing list at this index.
	for (table_hash=table->table[idx]; table_hash!=NULL; table_hash = table_hash->next) {
		if (table->compfn(table_hash->item, item)==0) {
			// collision, increment count
			return table_hash->magic;
		}
	}

	// Ok, found no collisions (this one doesn't exist). insert into list at head.
	// Could also sort these lists, helpful if they get long...
	item_hash = (struct hashtable_item *)local_new(sizeof(struct hashtable_item));
	item_hash->item = item;
	item_hash->magic = magic;
	item_hash->next = table->table[idx]; // null for first elem, ok.
	table->table[idx] = item_hash;
	table->nitems++;
	return magic;
}


// =============================================================================
// clear the hashtable, return an array of all the hashed items and all their magic.
// =============================================================================
int hash_unhash(struct hashtable *table, void ***pitems, int **pcounts) {
	void **items;
	int *counts;
	int count, hash_idx, table_idx;

	// Create the array to hold our data.
	if ((count = table->nitems)==0) {
		return 0;
	}
	items = (void **)local_new(count * sizeof(void *));

	if (counts!=NULL) {
		// ok, counts is not null. create an array to hold the data.
		counts=(int *)local_new(count*sizeof(int));
		// Then get the data out of the hashtable and into the array
		for (hash_idx=table_idx=0; table_idx<count; table_idx++) {
			if (hash_get_next(table, &(items[table_idx]), &hash_idx, &(counts[table_idx]))<0) {
				printf("Failed to get an item during exit (main)!\n");
				return -1;
			}
		}
		*pcounts = counts;
	} else {
		// Then get the data out of the hashtable and into the array
		// Same as code in other half of loop except for the NULL 'counts' paramter.
		for (hash_idx=table_idx=0; table_idx<count; table_idx++) {
			if (hash_get_next(table, &(items[table_idx]), &hash_idx, NULL)<0) {
				printf("Failed to get an item during exit (main)!\n");
				return -1;
			}
		}
	}

	*pitems = items;
	return count;
}


// =============================================================================
// print the table. pass in a callback to print a single element.
// =============================================================================
int print_table(FILE *file, struct hashtable *table) {
	int idx, printed;
	struct hashtable_item *item;

	if (table->nitems==0) {
		fprintf(file, "Table: EMPTY\n");
		return 1;
	}

	fprintf(file, "Table:\n");
	for (idx=0,printed=0; printed<table->nitems && idx<table->size; idx++) {
		if (table->table[idx]) {
			fprintf(file, "\t%d: ", idx);
			for (item=table->table[idx]; item; item=item->next) {
				(void)table->printfn(file, item->item);
				fprintf(file, "(%d)", item->magic);
				if (item->next) {
					fprintf(file, ", ");
				}
			}
			fprintf(file, "\n");
		}
	}

	return 1;
}


// =============================================================================
// Found on web, slightly modified.
// Do a web search for "g_str_hash X31_HASH" if you want to know more.
// =============================================================================
unsigned int string_hash(void *item, int size) {
	char *str;
	unsigned int hash;
	int len;

	str = (char *)item;
	len = strlen(str);

	for (hash=0; len; len--, str++) {
		hash = 31 * hash + *str;
	}

	return (hash%size);
}


// =============================================================================
// Xor first and second 16 bits, divide by 65537 (a prime, just over 65536)
//
// TODO Successively fold IP address into greatest number of power-of-two bits 
// TODO less than 'size'...
// =============================================================================
unsigned int ipaddr_hash(void *item, int size) {

	// the void pointer should be the IP address itself (it's only 32 bits).
	// This will xor the upper and lower halves, hashing it into 16 bits (64K) max.
	return(((((unsigned int)item)>>16) ^ (((unsigned int)item)&0xffff))%size);
}


// =============================================================================
// Compare two ip addresses, semantics of memcmp
// =============================================================================
int ipaddr_cmp(const void *item_a, const void *item_b) {
	// the void pointer should be the IP address itself (it's only 32 bits).
	// This will xor the upper and lower halves, hashing it into 16 bits (64K) max.
	return((unsigned int)item_a - (unsigned int)item_b);
}


// =============================================================================
// Print an IP address (dumb function).
//
// TODO: reference this code to print.h???
// =============================================================================
int ipaddr_print(FILE *file, const void *item) {
	// the void pointer should be the IP address itself (it's only 32 bits).
	// This will xor the upper and lower halves, hashing it into 16 bits (64K) max.
	return(fprintf(file, "%ld", (unsigned long)item));
}


// =============================================================================
// XOR the highest order bytes into the top two bytes of a 4-byte integer.
// then just do a stupid modulo.
// =============================================================================
unsigned int macaddr_hash(void *item, int size) {
	unsigned int hashval;
	unsigned char *p=(unsigned char*)item;

	hashval = (((unsigned int)(p[0]^p[2]))<<24) |
			  (((unsigned int)(p[1]^p[3]))<<16) |
			  (((unsigned int)(p[4]     ))<< 8) |
			  (((unsigned int)(p[5]     ))    );
	return (hashval%size);
}


// =============================================================================
// Compare two mac addresses; semantics of memcmp...
// =============================================================================
int macaddr_cmp(const void *item_a, const void *item_b) {
	return (memcmp(item_a, item_b, 6));
}


// =============================================================================
// Print a MAC address.
//
// TODO: reference this code to print.h???
// =============================================================================
int macaddr_print(FILE *file, const void *item) {
	unsigned char *p=(unsigned char*)item;

	return(fprintf(file, "%02X:%02X:%02X:%02X:%02X:%02X", p[0], p[1], p[2], p[3], p[4], p[5]));
}

