/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Anonymization code.
// #############################################################################
#include <errno.h>      // used when trying to mkdir
#include <fcntl.h>      // open()
#include <netinet/in.h> // ntohs
#include <stdio.h>      // printf.
#include <stdlib.h>     // open.
#include <string.h>     // strncpy, strncmp
#include <strings.h>    // bzero() and others
#include <sys/stat.h>   // open(), open flags
#include <sys/types.h>  // open()
#include <unistd.h>     // write.
#include "read_common.h"
#include "read_anon.h"
#include "print.h"
#include "hashtable.h"
#include "../common/shared_addresses.h"

// =============================================================================
// globals.
// =============================================================================
unsigned long cpu_speed, trace_start;
unsigned long long zero_time=0;
// for MAC anonymization.
static struct hashtable anon_mac;
static unsigned int anon_next_mac;
// for IP anonymization.
static struct hashtable anon_ip;
static unsigned int anon_next_ip;
// for port anonymization.
static unsigned short *anon_port;
static unsigned short anon_next_port;

// TODO: combine the next three functions into one generic get_anon_value function?
// TODO: may make reversal easier.

// =============================================================================
// Get the anonymized MAC ADDRESS for the given real MAC ADDRESS.
// Create a new one if it doesn't exist.
// =============================================================================
static unsigned int get_anon_mac(unsigned char *real_mac) {
	unsigned int anon_cur_mac;

	// add the given item to the table, if it exists.
	// returns: -1 on failure, index on success
	anon_cur_mac = hash_insert(&anon_mac, (void *)real_mac, anon_next_mac);
	if (anon_cur_mac==anon_next_mac) {
		anon_next_mac++;
	}
	return anon_cur_mac;
}


// =============================================================================
// Get the anonymized IP ADDRESS for the given real IP ADDRESS.
// Create a new one if it doesn't exist.
// =============================================================================
static unsigned int get_anon_ip(unsigned int real_ip) {
	unsigned int anon_cur_ip;

	// add the given item to the table, if it exists.
	// returns: -1 on failure, index on success
	anon_cur_ip = hash_insert(&anon_ip, (void *)real_ip, anon_next_ip);
	if (anon_cur_ip==anon_next_ip) {
		anon_next_ip++;
	}
	return anon_cur_ip;
}


// =============================================================================
// Get the anonymized PORT for the given real PORT.
// Create a new one if it doesn't exist.
// =============================================================================
static unsigned short get_anon_port(unsigned short real_port) {
	if (anon_port[real_port]==0) {
		anon_port[real_port]=anon_next_port++;
	}

	return anon_port[real_port];
}


// =============================================================================
// Open the output file, handles striping stuff.
// =============================================================================
static void open_outfile(char *in_file, int *outfd) {
	char out_file_name[512];
	int idx;

	if (*outfd != 0) {
		close(*outfd);
	}

	// old way -q
	// sprintf(out_file_name, "%s.anon", in_file);

	// Chomp out filename from path, and put in '/space/'
	for (idx=strlen(in_file); in_file[idx]!='/' && idx>=0; idx--) {
		;
	}
	idx++;
	sprintf(out_file_name, "/space/%s", &(in_file[idx]));

	// open
	if ((*outfd = open(out_file_name, O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))<0) {
		printf("Failed to open output file '%s' for input '%s'\n", out_file_name, in_file);
		exit(1);
	}
}


// =============================================================================
// Anonymize a single packet.
// =============================================================================
int anonymize_packet(struct collected_headers *hdr, int outfd) {
	struct processed_headers anon;

	memset((void *)&anon, '\0', sizeof(struct processed_headers));

	// convert to host endian-ness
	ntoh_collected_header(hdr);

	// Timestamp & length first.
	hdr->timestamp-=zero_time;
	anon.timestamp.tv_sec = (hdr->timestamp/cpu_speed) + trace_start;
	anon.timestamp.tv_nsec = (((double)(hdr->timestamp%cpu_speed))/cpu_speed)*1000000000;
	anon.length = hdr->length;

	// then MAC level stuff.
	anon.mac_src = get_anon_mac(hdr->mac_src);
	anon.mac_dst = get_anon_mac(hdr->mac_dst);
	anon.toslen = hdr->toslen;

	// Then IP level stuff.
	if (hdr->toslen == 0x0800) {
		anon.ip_src = get_anon_ip(hdr->ip_src);
		anon.ip_dst = get_anon_ip(hdr->ip_dst);
		anon.ip_len = hdr->ip_len;
		anon.ip_hdr_len = hdr->ip_hdr_len;
		anon.protocol= hdr->protocol;

		// then TCP/UDP level stuff.
		if (hdr->protocol==UDP_HDR_PROTOCOL || hdr->protocol==TCP_HDR_PROTOCOL) {
			anon.src_port = get_anon_port(hdr->src_port);
			anon.dst_port = get_anon_port(hdr->dst_port);

			// length is more complicated. Want to save the size of the data area.
			// For UDP, length of the packet (upper_len.udp_len) minus the 8-byte header.
			// for TCP, length of the packet (ip_len) minus variable header length (upper_len.tcp_len.len).
			if (hdr->protocol == UDP_HDR_PROTOCOL) {
				anon.upper_len = hdr->upper_len.udp_len-8;
			} else {
				anon.upper_len = hdr->ip_len - hdr->ip_hdr_len - hdr->upper_len.tcp_len.len;
			}
		}
	}

	// ----------------------------------------
	// output!
	// ----------------------------------------
	hton_processed_header(&anon);
	if (write(outfd, &anon, sizeof(struct processed_headers))!=sizeof(struct processed_headers)) {
		close(outfd);
		return 0;
	}

	return 1;
}


// =============================================================================
// Anonymization function that does mac/ip/ports.
//
// pkts delivered in order, cur_file is where they're currently being read from.
// =============================================================================
static int tcp_anon(struct dump_packet *pkt, char *cur_file) {
	static char *loc_in_file=NULL;
	static int outfd=0;
	unsigned int idx;

	// Give stats every so often (every 10000 packets)
	if (pkt->seqno%10000==0) {
		printf("Good sequence number %lld\n", pkt->seqno);
		fflush(stdout);
	}

	// ----------------------------------------
	// handle switching output files.
	// ----------------------------------------
	if (loc_in_file != cur_file) {
		printf("Detected filename change! %s --> %s\n", loc_in_file, cur_file);
		// close old file, open new for output.
		loc_in_file = cur_file;
		open_outfile(cur_file, &outfd);

		// ----------------------------------------
		// First-call initialization.
		// TODO: Do this elsewhere???
		// ----------------------------------------
		if (zero_time==0) {
			zero_time = pkt->data[0].timestamp;

			// create MAC anonymization table. Should only be a couple, the input/output firewall stuff.
			anon_next_mac = 1;
			if (hash_create(&anon_mac, 251, &macaddr_hash, &macaddr_cmp, &macaddr_print)<0) {
				printf("Failed to create new hash table for IP addresses!\n");
				return 0;
			}

			// create IP anonymization table.
			anon_next_ip = 1;
			if (hash_create(&anon_ip, 65536, &ipaddr_hash, &ipaddr_cmp, &ipaddr_print)<0) {
				printf("Failed to create new hash table for IP addresses!\n");
				return 0;
			}

			// create port anonymization table.
			anon_next_port = 1;
			if ((anon_port = (unsigned short *)calloc(65536, sizeof(unsigned short)))==NULL) {
				printf("Failed to allocate port number hash!\n");
				return 0;
			}
		}
	}

	// ----------------------------------------
	// Get copy of proper data and anonymize.
	// ----------------------------------------
	for (idx=0; idx<DUMPS_PER_PACKET; idx++) {
		if (!anonymize_packet(&(pkt->data[idx]), outfd)) {
			return 0;
		}
	}

	return 1;
}


// =============================================================================
// The main routine
// =============================================================================
int main(int argc, char *argv[]) {
	int idx, file_count=0;
	char **filenames=NULL;

	// parse command line.
	parse_reader_command_line(argc, argv, &file_count, &filenames, &cpu_speed, &trace_start);

	do_raw_read(filenames, file_count, &tcp_anon);

	// clean up allocations
	for (idx=0; idx<file_count; idx++) {
		free(filenames[idx]);
	}
	free(filenames);

	exit(0);
}

