/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Reader code; takes output saved to disk by writer, taken from kernel dump.
// #############################################################################
#include <netinet/in.h>       // ntohs, ntohl
#include <errno.h>            // used when trying to mkdir
#include <fcntl.h>            // open()
#include <stdlib.h>           // malloc(), close()
#include <string.h>           // strncpy, strncmp
#include <strings.h>          // bzero() and others
#include <unistd.h>           // close()
#include <stdio.h>
#include "read_common.h"
#include "../common/doubly.h"


// =============================================================================
// Usage: how to call this program.
// =============================================================================
void reader_usage(char *error) {
	if (error) { printf("\nUSAGE ERROR: %s\n", error); }
	printf("USAGE: ./reader [-c=cpu_speed]  [-s=trace_start] file1 [file2 ...]\n");
	printf("    * optional CPU speed is in hertz; try this:\n");
	printf("      `cat /proc/cpuinfo|grep 'cpu MHz'|sed -e 's/.* \\(.*\\)/\\1 100000*p/'|dc`\n");
	printf("    * optional trace start time is in seconds since the UNIX epoch;\n");
	printf("        you can get this by looking at the close time of a prior file,\n");
	printf("        or the system_info.txt file created by the writer program.\n");

	exit(1);
}


// =============================================================================
// Parse the command line.
// =============================================================================
void parse_reader_command_line(int argc, char *argv[], int *file_count, char ***filenames, unsigned long *cpu_speed, unsigned long *trace_start) {
	int idx, length;

	if (argc<2) {
		reader_usage("Too few arguments.");
	}

	// then get files list.
	*file_count = argc-1;

	if ((*file_count)<=0) {
		reader_usage("Failed to pass in file name[s] on command line.\n");
	}

	// get copy of all filenames.
	if ((*filenames = (char **)malloc((*file_count)*sizeof(char *)))==NULL) {
		perror("Failed to allocate file name pointers!\n");
		exit(1);
	}
	for ((*file_count)=0,idx=1; idx<argc; idx++) {
		if (argv[idx][0] != '-') {
			// Not an option. Add filename to list.
			length = strlen(argv[idx]) + 1;
			if (((*filenames)[*file_count] = (char *)malloc((length)*sizeof(char)))==NULL) {
				perror("Failed to allocate file name pointers!\n");
				exit(1);
			}
			memcpy((*filenames)[*file_count], argv[idx], length);
			(*file_count)++;
		} else {
			// TODO???? Verbosity handling here.
			// an option. parse it out.
			if (strlen(argv[idx])<4) {
				reader_usage("Short option!\n");
			}
			if ((length = atol(&(argv[idx][3])))<1000000) {
				reader_usage("Bad option (<1 million)");
			}

			// save it to parameter.
			switch(argv[idx][1]) {
				case 'c': case 'C': if (cpu_speed  !=NULL) { *cpu_speed   = length; } break;
				case 's': case 'S': if (trace_start!=NULL) { *trace_start = length; } break;
				default: reader_usage("Bad option on command line!");
			}
		}
	}
}


// =============================================================================
// Perform endian conversions on all fields in our collected headers.
// =============================================================================
void ntoh_collected_header(struct collected_headers *hdr) {
	// hdr->length= ntohs(hdr->length); // OOPs! length is already little-endian.
	hdr->toslen   = ntohs(hdr->toslen);
	hdr->ip_src   = ntohl(hdr->ip_src);
	hdr->ip_dst   = ntohl(hdr->ip_dst);
	hdr->ip_len   = ntohs(hdr->ip_len);
	hdr->src_port = ntohs(hdr->src_port);
	hdr->dst_port = ntohs(hdr->dst_port);

	// This weird conversion is because the header length
	// is really in 32-bit words. The collection code copied it
	// from the last 4 bits of the packet field into the first char of
	// the local field. Convert to bytes by multiplying it by 4.
	// 20020607: removed, already done by kernel code.
	// hdr->ip_hdr_len = ((hdr->ip_hdr_len)&0xF)<<2;

	// the "upper_len" field is different for TCP and UDP
	if (hdr->protocol==UDP_HDR_PROTOCOL) {
		// packet length in NBO. convert to HBO.
		hdr->upper_len.udp_len = ntohs(hdr->upper_len.udp_len);
	} else if (hdr->protocol==TCP_HDR_PROTOCOL) {
		// header length, in 32-bit words. Convert to bytes via shift left 2 or multiply by 4
		hdr->upper_len.udp_len = ((unsigned short)hdr->upper_len.tcp_len.len) << 2;
		// NOTE: I've seen some odd things with this next line of code on some packets;
		// Some incongruity between header lengths screws up here. I haven't tracked it down
		// quite yet... Assuming it's a buggy TCP/IP stack of some sort, but not sure.
		// Then, we really want the actual packet length, not the length of the header (like UDP)
		// In short, I saw negative or too-small numbers.
		//     hdr->upper_len.udp_len = hdr->ip_len - hdr->upper_len.udp_len;
	}
}


// =============================================================================
// Perform endian conversions on all fields in our collected headers.
// =============================================================================
void ntoh_processed_header(struct processed_headers *hdr) {
	hdr->timestamp.tv_sec  = ntohl(hdr->timestamp.tv_sec);
	hdr->timestamp.tv_nsec = ntohl(hdr->timestamp.tv_nsec);
	hdr->length            = ntohs(hdr->length);

	hdr->mac_src = ntohl(hdr->mac_src);
	hdr->mac_dst = ntohl(hdr->mac_dst);
	hdr->toslen  = ntohs(hdr->toslen);

	hdr->ip_src = ntohl(hdr->ip_src);
	hdr->ip_dst = ntohl(hdr->ip_dst);
	hdr->ip_len = ntohs(hdr->ip_len);

	hdr->src_port  = ntohs(hdr->src_port);
	hdr->dst_port  = ntohs(hdr->dst_port);
	hdr->upper_len = ntohs(hdr->upper_len);
}


// =============================================================================
// Perform endian conversions on all fields in our collected headers.
// =============================================================================
void hton_processed_header(struct processed_headers *hdr) {
	hdr->timestamp.tv_sec  = htonl(hdr->timestamp.tv_sec);
	hdr->timestamp.tv_nsec = htonl(hdr->timestamp.tv_nsec);
	hdr->length            = htons(hdr->length);

	hdr->mac_src = htonl(hdr->mac_src);
	hdr->mac_dst = htonl(hdr->mac_dst);
	hdr->toslen  = htons(hdr->toslen);

	hdr->ip_src = htonl(hdr->ip_src);
	hdr->ip_dst = htonl(hdr->ip_dst);
	hdr->ip_len = htons(hdr->ip_len);

	hdr->src_port  = htons(hdr->src_port);
	hdr->dst_port  = htons(hdr->dst_port);
	hdr->upper_len = htons(hdr->upper_len);
}


// =============================================================================
// Compare packet sequence numbers; for sorting.
// =============================================================================
static int pkt_seqno_compare(void *pkt_a, void *pkt_b) {
	return (((struct dump_packet *)pkt_b)->seqno - ((struct dump_packet *)pkt_a)->seqno);
}


// =============================================================================
// read packets from the file. For driving anonymization or analysis code.
//
// This is basically the producer/consumer generic code.
//     producer --> read packets into sorted list.
//     consumer --> then take them out and analyzes them.
// =============================================================================
void do_raw_read(char **filenames, int file_count, int (*analysis_fn)(struct dump_packet *pkt, char *cur_file)) {
	struct dump_packet *cur_pkt;
	int idx, infd;
	unsigned long long nread=0, next_seqno=0;
	struct dlinked_list pkt_list;
	memset(&pkt_list, '\0', sizeof(struct dlinked_list));
	pkt_list.compfn = &pkt_seqno_compare;

	// read all files.
	for (idx=0; idx<file_count; idx++) {
		printf("Reading input file '%s'\n", filenames[idx]);

		// open file.
		if ((infd = open(filenames[idx], O_RDONLY))<0) {
			printf("FAILURE! Can't open input file '%s' because '%s'\n", filenames[idx], strerror(errno));
			exit(1);
		}

		// Ready to read
		while (1) {
			// produce packets.
			if ((cur_pkt=(struct dump_packet *)malloc(sizeof(struct dump_packet)))==NULL) {
				printf("Couldn't create dump packet input ring index %d!\n", idx);
				exit(1);
			}
			if (read(infd, cur_pkt, sizeof(struct dump_packet))==sizeof(struct dump_packet)<=0) {
				// break loop when file completed.
				break;
			}
			nread++;

			// silently drop packets that are duplicated or so badly out of sequence that
			// we've already declared them LOST.
			if (cur_pkt->seqno >= next_seqno) {
				dlist_insert(&pkt_list, (void *)cur_pkt);
			}

			// handle loss.
			if (pkt_list.count>128) {
				// TODO: silently drop?????
				// printf("Packet lost (outside of window)!\n");
				dlist_pop_head(&pkt_list, (void **)&cur_pkt);
				next_seqno = cur_pkt->seqno;
				dlist_push_head(&pkt_list, (void *)cur_pkt);
			}

			// consume packets.
			while (pkt_list.count>0) {
				// get packet, see if it's what we want.
				dlist_pop_head(&pkt_list, (void **)&cur_pkt);
				if (cur_pkt->seqno==next_seqno) {
					// yup! analyze.
					if (!(*analysis_fn)(cur_pkt, filenames[idx])) {
						goto end;
					}
					free(cur_pkt);
					next_seqno++;
				} else {
					// printf("Out-of-order packet, wanted %lld got %lld\n", next_seqno, cur_pkt->seqno);
					// nope. Save for next time. Handle case of *really* out-of-order packets.
					dlist_push_head(&pkt_list, (void *)cur_pkt);
					break;
				}
			}
		}
		// close files
		close(infd);

		// flush rest of list
		while (pkt_list.count>0) {
			dlist_pop_head(&pkt_list, (void **)&cur_pkt);
			if (!(*analysis_fn)(cur_pkt, filenames[idx])) {
				goto end;
			}
			free(cur_pkt);
		}

		// give stats.
		// printf("After having read file '%s', nread is %lld\n", filenames[idx], nread);
	}

end:
	close(infd);
	return;
}


// =============================================================================
// read processed data from the file. For driving analyzer code.
// =============================================================================
void do_processed_read(char **filenames, int file_count, size_t struct_size, int (*analysis_fn)(void *data, char *cur_file)) {
	void *data;
	int idx, infd;
	unsigned long long nread=0;

	if ((data = calloc(1, struct_size))==NULL) {
		printf("Failed allocation of %d bytes in do_processed_read()\n", struct_size);
		exit(1);
	}

	// read all files.
	for (idx=0; idx<file_count; idx++) {
		// open file.
		if ((infd = open(filenames[idx], O_RDONLY))<0) {
			printf("FAILURE! Can't open input file '%s' because '%s'\n", filenames[idx], strerror(errno));
			exit(1);
		}

		// Ready to read
		for(; read(infd, data, struct_size)==struct_size; nread++) {
			// ----------------------------------------
			// put your analysis code here.
			// ----------------------------------------
			analysis_fn(data, filenames[idx]);
		}
		// close files
		close(infd);

		// give stats.
		printf("After having read file '%s', nread is %lld\n", filenames[idx], nread);
	}
}

