/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// A simple program to read in the first and last records from a file, and to
// print out their sequence number and timestamp.
//
// Useful to find out how much data in each file, and what time range each
// file holds.
// #############################################################################
#include <stdio.h>
#include <fcntl.h>            // open()
#include <errno.h>            // used when trying to mkdir
#include <string.h>           // strncpy, strncmp
#include <strings.h>          // bzero() and others
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include "read_common.h"
#include "print.h"

// =============================================================================
// The main routine
// =============================================================================
int main(int argc, char *argv[]) {
	int infd, idx, file_count=0;
	char **filenames=NULL;
	unsigned long cpu_speed=0;
	struct dump_packet first_pkt, last_pkt;
	struct stat stat_buf;
	off_t offset;

	// parse command line.
	parse_reader_command_line(argc, argv, &file_count, &filenames, &cpu_speed, NULL);

	// read all files.
	for (idx=0; idx<file_count; idx++) {
		printf("Reading input file '%s'\n", filenames[idx]);

		// open file.
		if ((infd = open(filenames[idx], O_RDONLY))<0) {
			printf("FAILURE! Can't open input file '%s' because '%s'\n", filenames[idx], strerror(errno));
			exit(1);
		}

		if (fstat(infd, &stat_buf)<0) {
			printf("Failed to stat '%s' because '%s'\n", filenames[idx], strerror(errno));
			close(infd);
			exit(1);
		}
		if (stat_buf.st_size < sizeof(struct dump_packet)) {
			printf("Cannot read file; size too small. (%lu < %lu)\n", stat_buf.st_size, (unsigned long)sizeof(struct dump_packet));
			close(infd);
			exit(1);
		}

		// Read first and last packets.
		if (read(infd, &(first_pkt), sizeof(struct dump_packet))==sizeof(struct dump_packet)<=0) {
			printf("Couln't read first packet!\n");
			exit(1);
		}

		// This seek doesn't work correctly on abruptly terminated files.
		if ((stat_buf.st_size % sizeof(struct dump_packet))!=0) {
			printf("Warning: file size (%lu) is not multiple of size of a dump packet (%lu)!\n", stat_buf.st_size, (unsigned long)sizeof(struct dump_packet));
		}

		offset = -1 * (sizeof(struct dump_packet)+ (stat_buf.st_size%sizeof(struct dump_packet)));
		if (lseek(infd, offset, SEEK_END)<offset) {
			printf("Couln't seek to end!\n");
			exit(1);
		}

		if (read(infd, &(last_pkt), sizeof(struct dump_packet))==sizeof(struct dump_packet)<=0) {
			printf("Couln't read last packet!\n");
			exit(1);
		}

		// print first and last packets.
		if (cpu_speed>0) {
			printf("%lld %f\n", first_pkt.seqno, ((double)first_pkt.data[0].timestamp/(double)cpu_speed));
			printf("%lld %f\n", last_pkt.seqno, ((double)last_pkt.data[DUMPS_PER_PACKET-1].timestamp/(double)cpu_speed));
		} else {
			printf("%lld %lld\n", first_pkt.seqno, first_pkt.data[0].timestamp);
			printf("%lld %lld\n", last_pkt.seqno, last_pkt.data[DUMPS_PER_PACKET-1].timestamp);
		}

		// close files
		close(infd);
	}

	// clean up allocations
	for (idx=0; idx<file_count; idx++) {
		free(filenames[idx]);
	}
	free(filenames);

	exit(0);
}

