/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Verify validity of packet data using redundancy checks.
//
// TODO: for checking of fragments... eventually.
// #############################################################################
#include <fcntl.h>     // needed for open, too.
#include <stdio.h>
#include <stdlib.h>     // free
#include <sys/stat.h>   // open(), open flags()
#include <sys/types.h>  // needed for open, too.
#include <unistd.h>     // write
#include "read_common.h"
#include "print.h"

int verify_data;
static int verifyfd;


// =============================================================================
// Verify the packets in struct dump_packet.
//
// Cur_file (name of file from which this dump_packet came) is unused here.
//    just need it because of function pointer syntax.
// =============================================================================
int verify_packet(struct dump_packet *pkt, char *cur_file) {
	int idx;
	struct collected_headers *hdr;

	for (idx=0; idx<DUMPS_PER_PACKET; idx++) {
		hdr = &(pkt->data[idx]);

		ntoh_collected_header(hdr);

		if (hdr->toslen == 0x0800) {
			// IP traffic...

			// Minimum packet size on ethernet is 64 (14 hdr + 4 chksum at lowest-level);
			// so if data segment is less than 46 bytes it's padded with zeros.
			// In this trace, the padding occurs frequently with GRE and ICMP packets.
			if ((hdr->length - 14)!=hdr->ip_len && hdr->length>60 /* don't see 4-byte CRC */) {
				printf("Failed IP length redundancy check: ");
				print_data_tcpdump(stdout, hdr);
			}

			// IP headers generally do not have options.
			if (hdr->ip_hdr_len!=20) {
				printf("Unusually-sized IP header: %d\n", hdr->ip_hdr_len);
				print_data_tcpdump(stdout, hdr);
			}
			if (hdr->protocol==UDP_HDR_PROTOCOL || hdr->protocol==TCP_HDR_PROTOCOL) {
				if (hdr->protocol == UDP_HDR_PROTOCOL) {
					// UDP header lengths are redundant; check 'em.
					if ((hdr->ip_len - hdr->ip_hdr_len)!=hdr->upper_len.udp_len) {
						// printf("Failed UDP length redundancy check: ");
						// print_data_tcpdump(stdout, hdr);
						// Just save in HBO.
						if (write(verifyfd, hdr, sizeof(struct collected_headers))!=sizeof(struct collected_headers)) {
							close(verifyfd);
							verifyfd=0;
							return 0;
						}
					}
				} else {
					// We only expect certain TCP header option lengths: 20 bytes ... 60 bytes, must be multiple of 4.
					if (hdr->upper_len.tcp_len.len%4!=0 || hdr->upper_len.tcp_len.len<20 || hdr->upper_len.tcp_len.len>60) {
						printf("Funny-sized TCP packet: ");
						print_data_tcpdump(stdout, hdr);
						return 0;
					}
				}
			}
		} else if (hdr->toslen!=0x0806) {
			// ARP traffic...

			// We should only see IP and ARP traffic; other traffic probably a corrupt packet.
			printf("Non-IP, non ARP packet: ");
			print_data_tcpdump(stdout, hdr);
			return 0;
		}
	}

	return 1;
}


// =============================================================================
// The main routine
// =============================================================================
int main(int argc, char *argv[]) {
	int idx, file_count=0;
	char **filenames=NULL;

	// parse command line.
	parse_reader_command_line(argc, argv, &file_count, &filenames, NULL, NULL);

	if ((verifyfd = open("/space/fragment.verify", O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))<0) {
		printf("Failed to open fragment verification file for output!\n");
		exit(1);
	}

	do_raw_read(filenames, file_count, &verify_packet);

	close(verifyfd);

	// clean up allocations
	for (idx=0; idx<file_count; idx++) {
		free(filenames[idx]);
	}
	free(filenames);

	exit(0);
}

