/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Shower program.  shows dump packets as they come across the
//    wire but doesn't save them like the writer.
//
// For debugging, mostly.
//
// TODO: integrate common parts between shower and writer.
// TODO: Perhaps use code in common/user_misc.h to do timestamping &c.
// #############################################################################
#include "dump_common.h"

// =============================================================================
// GLOBAL VARIABLES
// =============================================================================
// Command line parameters
UINT64  packet_count=0;
int     sockfd=0, verbosity;


// =============================================================================
// LOCAL FUNCTION PROTOTYPES
// =============================================================================
// Tell the user how they screwed up ;)
void usage(char *msg);
// Function called on ^C
void terminate(int var);
// parse command-line options
void parseCmdLine(int argc, char *argv[]);
// get the timestamp.
unsigned long long get_timestamp();
// open and configure our socket.
void open_socket(struct sockaddr_in *sock_addr, int *sockfd);


// =============================================================================
// how to call program.
// =============================================================================
void usage(char *msg) {
	printf("USAGE ERROR: '%s'\n", msg);
	printf("USAGE: ./shower <-v> <-v>...\n");
	printf("    -v: run verbosely\n");
	exit(1);
}


// =============================================================================
// get command line
// =============================================================================
void parseCmdLine(int argc, char *argv[]) {
	int i;

	// initialize variables.
	verbosity=0;

	// Fill in options fields.
	for (i=1; i<argc; i++) {
		// start checking input against each flag.
		if (strlen(argv[i])==2) {
			if (argv[i][0]!='-') {
				usage("Bad flag delimiter");
			}
			switch (argv[i][1]) {
				case 'v': verbosity++; break;
				default:  usage("Bad flag"); break;
			}
		} else {
			usage("Bad option length");
		}
	}
}


// =============================================================================
// The main routine
// =============================================================================
int main(int argc, char *argv[]) {
	struct sigaction my_sigact;
	struct dump_packet packet;
	struct sockaddr_in sock_addr;
	int n, addr_len;

	// parse command line.
	parseCmdLine(argc, argv);

	// Handle termination cleanly.
	bzero(&my_sigact, sizeof(struct sigaction));
	my_sigact.sa_handler = &(terminate);
	if (sigaction(SIGINT, &my_sigact, NULL)) {
		perror("Could not set up signal handlers.\n");
		return 1;
	}

	open_socket(&sock_addr, &sockfd);

	// now tell the user they've started up a server OK.
	printf("Ready to receive dump packets...\n");

	// now loop for messages from client (requests)
	packet_count=0;
	read_loop:
		// read a packet
		addr_len = sizeof(sock_addr);
		if ((n=recvfrom(sockfd, &packet, sizeof(packet), MSG_WAITALL, 
				(struct sockaddr *) &sock_addr, (socklen_t *) &addr_len))
				!= sizeof(struct dump_packet)) {
			printf("Error %d; Received partial packet (length %d, not %d)!\n", errno, n, sizeof(struct dump_packet));
			exit(1);
		}

		// tell user
		if (verbosity>0) {
			printf(".");
		}
		// save to disk.
		printf("Got a packet!\n");

		packet_count++;
	goto read_loop;

	exit(0);
}

// =============================================================================
// server's shutdown routine, run asynchronously.
// =============================================================================
void terminate(int sig) {
	printf("Received %llu packets total.\n", packet_count);
	close(sockfd);
	exit(0);
}


// =============================================================================
// get the timestamp.
// =============================================================================
unsigned long long get_timestamp() {
	register unsigned int dummylo;
	register unsigned int dummyhi;
	unsigned long long dummy;
	
	asm volatile (".byte 0x0f, 0xa2" : : : "eax", "ebx", "ecx", "edx");
	asm volatile (".byte 0x0f, 0x31;              # the RDTSC instruction
				  movl %%eax, %0 
				  movl %%edx, %1" 
				  : "=r" (dummylo), "=r" (dummyhi)
				  :
				  : "eax", "edx"
				  );

	dummy = dummyhi;
	dummy = (dummy << 32) | dummylo;

	return dummy;
}


// =============================================================================
// open and configure our socket.
// =============================================================================
void open_socket(struct sockaddr_in *sock_addr, int *sockfd) {
	int size;

	/* Opens a UDP socket (an internet datagram socket) */
	if ((*sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("Dumper(socket)");
		exit(1);
	}
	if (verbosity>0) {
		printf("Called socket...\n");
	}

	/* Bind our local address so that the client can send to us */
	bzero(sock_addr, sizeof(*sock_addr));
	sock_addr->sin_family = AF_INET;
	sock_addr->sin_addr.s_addr = htonl(INADDR_ANY);
	sock_addr->sin_port = htons(CLIENT_DUMP_PORT);
	if (bind(*sockfd, (struct sockaddr *) sock_addr, sizeof(*sock_addr)) != 0) {
		perror("Dumper(bind)");
		exit(1);
	}
	if (verbosity>0) {
		printf("Bound socket appropriately.\n");
	}

	// Try to set up the socket receive buffer size; 64MB
	size = 64 * 1024 * 1024 * 4; // 67108864
	if (setsockopt(*sockfd, SOL_SOCKET, SO_RCVBUF, &size, 4)!=0) {
		printf("Failed to set socket option (64MB receive buffer)\n");
		printf("Try `echo 67108864 > /proc/sys/net/core/rmem_max`\n");
		perror("Foo!");
		exit(1);
	}
	if (verbosity>0) {
		printf("Set socket receive buffer size to 64MB\n");
	}
}

