/*
###############################################################################
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
###############################################################################
*/
// #############################################################################
// Writer program. Saves data coming from kernel box to local disk.
// #############################################################################
#include "../common/user_network.h"
#include "../common/shared_dump.h"
#include "../common/user_misc.h"
#include "../common/config.h"
#include <signal.h>           // sigaction
#include <time.h>             // time.
#include <sys/stat.h>         // for mkdir
#include <sys/types.h>        // for mkdir, open()
#include <unistd.h>           // close(), sleep
#include <stdlib.h>           // malloc()
#include <sys/vfs.h>          // statfs

// ??? 20021023; removed this and changed its ONE use to unsigned long long packet_count...
// typedef unsigned long long UINT64;


// =============================================================================
// GLOBAL VARIABLES
// =============================================================================
// The C[++] standard may define these to be nulled out,
// but I like to do so myself.
unsigned long long packet_count=0;
int sockfd=0, verbosity=0;
int bufsize;
// For writing to files.
FILE *DUMPFILE=NULL;
char **devices=NULL;
int dev_idx=0;
// Scratch buffer used in various places.
int tmpbuf_size=0;
char *tmpbuf=NULL;


// =============================================================================
// LOCAL FUNCTIONS, declared
// =============================================================================
// Tell the user how they screwed up ;)
void usage(char *error);
// verify validity of command-line parameters
int check_devices(int count, char *device_names[]);
// parse command-line options
void parse_command_line(int argc, char *argv[]);
// Return the amount of space free on 'device_name'
int disk_free(char *device_name);
// Function called on ^C or other deadly event
void terminate(int sig);
// split the dump file when it gets to the flag size.
void dump_file_split(void);


// =============================================================================
// how to call program.
// =============================================================================
void usage(char *error) {
	if (error) { printf("\nUSAGE ERROR: %s\n", error); }
	printf("USAGE: ./writer <-v[vv...]> <-b=[bufsize]> [directory] <directory2...>\n\n");
	printf("    -v, -vv, -vvv: run verbosely, very verbosly, and in debugging mode\n");
	printf("    -b=[bufsize]: set the socket buffer size to this value.\n"); 
	printf("Pass in any number of directories in which to store output files;\n");
	printf("we write files to each directory until its device is full, then move on.\n");
	exit(1);
}


// =============================================================================
// Check that each passed-in parameter is actually on a separate device.
// If two or more are on the same device, the striping code fails (we
// muck with the first one until its full, then when we hit the second
// one the disk is already full).
// =============================================================================
int check_devices(int count, char *device_names[]) {
	struct stat *file_info;
	int idx, jdx;

	// allocate memory
	if ((file_info=(struct stat *)malloc(count*sizeof(struct stat)))==NULL) {
		printf("Error: Failed to allocate file_info structure!\n");
		return 0;
	}

	// first need to get info
	for (tmpbuf_size=idx=0; idx<count; idx++) {
		if (verbosity>1) {
			printf("Checking file at '%s'...\n", device_names[idx]);
		}
		// this code ensures that the temporary buffer is large enough to hold
		// what we'll use it for.
		if (strlen(device_names[idx])>tmpbuf_size) {
			tmpbuf_size = strlen(device_names[idx]);
		}

		if (stat(device_names[idx], &(file_info[idx]))!=0) {
			printf("Error: Failed to stat FILE '%s' because '%s'.\n", device_names[idx], strerror(errno));
			return 0;
		}
	}

	// create temporary buffer, nicely aligned. Include space for max pathname,
	// plus filename appended (see dump_file_split()'s sprintf).
	idx=tmpbuf_size+32;
	tmpbuf_size=1;
	while (tmpbuf_size<idx) {
		tmpbuf_size*=2;
	}
	if ((tmpbuf=(char *)malloc(tmpbuf_size))==NULL) {
		printf("Could not create scratch buffer!\n");
		return 0;
	}

	// compare all devices to ensure each is unique.
	for (idx=0; idx<(count-1); idx++) {
		for (jdx=(idx+1); jdx<count; jdx++) {
			if (memcmp((void *)&(file_info[idx].st_dev), (void *)&(file_info[jdx].st_dev), sizeof(dev_t))==0) {
				printf("Error: '%s' and '%s' are on same device.\n", device_names[idx], device_names[jdx]);
				printf("I can't figure out striping in this case\n");
				return 0;
			}
		}
	}
	if (verbosity>2) {
		printf("\tDevices look good; each is unique.\n");
	}

	// free memory
	free(file_info);

	// done!
	return 1;
}


// =============================================================================
// get command line parameters
// =============================================================================
void parse_command_line(int argc, char *argv[]) {
	int idx, len;

	// move past program name.
	argv = &(argv[1]);
	argc--;

	// then get verbosity
	if (argc>0 && argv[0][0]=='-' && (argv[0][1]=='v' || argv[0][1]=='V')) {
		verbosity=strlen(argv[0])-1;
		argv = &(argv[1]);
		argc--;
	} else {
		verbosity=0;
	}

	// then get buffer size
	if (argc>0 && argv[0][0]=='-' && (argv[0][1]=='b' || argv[0][1]=='B') && argv[0][2]=='=') {
		if (sscanf(&(argv[0][3]), "%d", &bufsize)!=1) {
			usage("Failed scan of bufsize string!");
		}
		if (bufsize<=0) {
			usage("Scanned bad (negative or zero) bufsize string!");
		}

		argv = &(argv[1]);
		argc--;
	} else {
		bufsize=0;
	}

	if (argc<1) { usage("Too few arguments"); }

	// legal input?
	if (!check_devices(argc, argv)) {
		usage("Parameters passed in are not on legal devices!");
	}

	// get copy of device names, and null-terminate
	if ((devices=(char **)malloc((argc)*sizeof(char **)))==NULL) {
		printf("Failed allocation of device_list copy\n");
		exit(1);
	}
	for (idx=0; idx<argc; idx++) {
		len = strlen(argv[idx])+1;
		if ((devices[idx]=(char *)malloc(len))==NULL) {
			printf("Failed allocation of device_list copy index %d\n", idx);
			exit(1);
		}
		memcpy(devices[idx], argv[idx], len);
	}
	devices[idx] = NULL;

	// Ok, give some info now.
	if (verbosity>0) {
		printf("File size will be %d bytes, taking up about %d blocks/file.\n", FILE_SIZE_BYTES, FILE_SIZE_BLOCKS);
		for (dev_idx=0; devices[dev_idx]!=NULL; dev_idx++) {
			printf("\t'%s' (device %d) will hold about %d files.\n",
					devices[dev_idx], dev_idx, disk_free(devices[dev_idx]));
		}
		// have to reset before continuing!
		dev_idx=0;
	}
}


// =============================================================================
// How much disk space (in terms of FILES) is left on the given device?
// See statfs(2) for some useful info.
//
// The arithmetic below uses integers (not long longs or doubles).
// On a 32-bit system it should work for disks up to 2TB in size.
//
// This algorithm has some problems, such as the fact that FILE_SIZE_BLOCKS
// is not really known beforehand. To account for this I add some fluff to
// the factors (+4k fluff in the FILE_SIZE_BLOCKS #define).
// =============================================================================
int disk_free(char *device_name) {
	struct statfs fs_info;

	// now get filesystem information.
	if (verbosity>2) {
		printf("Checking filesystem at '%s'\n", device_name);
	}
	if (statfs(device_name, &fs_info)!=0) {
		printf("Failed to stat '%s' because '%s'\n", device_name, strerror(errno));
		exit(1); // just return zero and 'soft fail'?
	}

	// root reserves some space, return appropriate value depending on whether we're root
	// or a user who cannot access that disk space.
	if (getuid()==0) {
		// printf("Root availability... %ld blocks, %ld files\n", fs_info.f_bsize * (fs_info.f_bfree/1024), (fs_info.f_bsize * (fs_info.f_bfree/1024))/FILE_SIZE_BLOCKS ); // 1KB blocks available to users
		return((fs_info.f_bsize * (fs_info.f_bfree/1024))/FILE_SIZE_BLOCKS - 1);
	} else {
		// printf("User availability... %ld blocks, %ld files\n", fs_info.f_bsize * (fs_info.f_bavail/1024), (fs_info.f_bsize * (fs_info.f_bavail/1024))/FILE_SIZE_BLOCKS ); // 1KB blocks available to users
		return((fs_info.f_bsize * (fs_info.f_bavail/1024))/FILE_SIZE_BLOCKS - 1);
	}

	return 0;
}


// =============================================================================
// server's shutdown routine, run asynchronously.
// =============================================================================
void terminate(int sig) {
	int idx;

	printf("Received %llu packets total.\n", packet_count);

	if (sockfd!=0) { close(sockfd); }
	if (tmpbuf!=NULL) { free(tmpbuf); }

	if (devices!=NULL) {
		for (idx=0; devices[idx]!=NULL; idx++) {
			free(devices[idx]);
		}
		free(devices);
	}
	exit(0);
}


// =============================================================================
// split the dump file when it gets to the flag size.
// =============================================================================
void dump_file_split(void) {
	static int abs_file_num=0;
	static int rel_file_num=0;

	// tell user we're splitting.
	if (verbosity>0) {
		printf("S(%d) ", abs_file_num);
	}

	// close last file if it exists.
	if (DUMPFILE!=NULL && fclose(DUMPFILE)!=0) {
		perror("Failure closing previous dumpfile!\n");
		exit(-1);
	}

	// Need to move to new mount point/device
	// Add some fluff here (2 instead of 1):
	//   this ensures we always leave space for an extra file per device.
	while (disk_free(devices[dev_idx])<2) {
		if (devices[dev_idx]==NULL || devices[dev_idx+1]==NULL) {
			printf("I've written %d files, filling all given devices. Terminating\n", abs_file_num);
			terminate(0);
		}
		dev_idx++;

		// Make the directory. see chmod(2) for permission info
		sprintf(tmpbuf, "%s/%03d/", devices[dev_idx], abs_file_num/DUMP_FILES_PER_DIR);
		if (mkdir(tmpbuf, S_IRWXU|S_IRWXG|S_IRWXO)<0) {
			if (errno != EEXIST) {
				printf("Could not make directory '%s' because '%s'\n", tmpbuf, strerror(errno));
				terminate(1);
			}
		}

		rel_file_num=0;
	}

	// move to next directory when this one full.
	if (!(abs_file_num%DUMP_FILES_PER_DIR)) {
		// Make the directory. see chmod(2) for permission info
		sprintf(tmpbuf, "%s/%03d/", devices[dev_idx], abs_file_num/DUMP_FILES_PER_DIR);
		if (mkdir(tmpbuf, S_IRWXU|S_IRWXG|S_IRWXO)<0) {
			if (errno != EEXIST) {
				printf("Could not make directory '%s' because '%s'\n", tmpbuf, strerror(errno));
				terminate(1);
			}
		}

		rel_file_num=0;
	}

	// open new file.
	sprintf(tmpbuf, "%s/%03d/file.%06d", devices[dev_idx], abs_file_num/DUMP_FILES_PER_DIR, abs_file_num);
	if ((DUMPFILE=fopen(tmpbuf, "w"))==NULL) {
		printf("Can't open dumpfile '%s'\n", tmpbuf);
		perror("Error was");
		exit(-1);
	}

	// increment file counter.
	abs_file_num++;
	rel_file_num++;
}


// =============================================================================
// Save system information including the time and CPU speed for post analysis.
// =============================================================================
void save_system_info(void) {
	FILE *status_file;

	sprintf(tmpbuf, "%s/system_info.txt", devices[0]);

	if ((status_file=fopen(tmpbuf, "w"))==NULL) {
		printf("WARNING! Failed to open system info file '%s'\n", tmpbuf);
		return;
	}
	fprintf(status_file, "The UNIX time is %lu.\n\n", time(NULL));
	fprintf(status_file, "Information from /proc/cpuinfo follows:\n");
	fclose(status_file);

	sprintf(tmpbuf, "cat /proc/cpuinfo >> %s/system_info.txt", devices[0]);
	system(tmpbuf);
}


// =============================================================================
// The main routine
// =============================================================================
int main(int argc, char *argv[]) {
	struct sigaction my_sigact;
	struct dump_packet packet;
	struct sockaddr_in serv_addr;
	int len, addr_len;

	// parse command line.
	parse_command_line(argc, argv);

	// Handle termination cleanly.
	bzero(&my_sigact, sizeof(struct sigaction));
	my_sigact.sa_handler = &(terminate);
	if (sigaction(SIGINT, &my_sigact, NULL)) {
		perror("Could not set up signal handlers.\n");
		return 1;
	}

	// network startup
	if (!(sockfd=create_udp_socket(NULL, &serv_addr, bufsize, verbosity))) {
		printf("Failed to create a socket!\n");
		exit(1);
	}

	// save cpu speed & time.
	save_system_info();

	if (verbosity>0) { printf("Ready to receive dump packets...\n"); }

	// now loop forever for messages from client
	while (1) {
		// move to next file
		dump_file_split();

		for (packet_count=0; packet_count<RECORDS_PER_FILE; packet_count++) {
			// read a packet
			addr_len = sizeof(serv_addr);
			if ((len=recvfrom(sockfd, &packet, sizeof(packet), MSG_WAITALL,
					(struct sockaddr *) &serv_addr, (socklen_t *) &addr_len))
					!= sizeof(struct dump_packet)) {
				printf("Error %s; Received partial packet (length %d, not %d)!\n", strerror(errno), len, sizeof(struct dump_packet));
			} else {
				// tell user
				if (verbosity>0) { printf("."); }

				// Actually write to the file.
				if (fwrite(&packet, sizeof(struct dump_packet), 1, DUMPFILE)!=1) {
					printf("Failed write to file!!!\n");
					perror("Error was");
					exit(-1);
				}

				// tell user we're writing.
				if (verbosity>0) { printf("W"); }
			}
		}
	}

	terminate(0);
}

